# This code connects to IBL back end and resolves
# all point neurons (channels) measured in all
# valid experiments.
try:
    from oneibl.one import ONE
except ImportError:
    # Then we're on ONE2
    from one.api import ONE
import os
import pickle
import numpy as np

import ibllib.atlas as atlas
from brainbox.io.one import load_channels_from_insertion
import alf.io

from iblviewer.mouse_brain import MouseBrainViewer
from iblviewer import utils


def get_valid_insertions(force_query=False, local_file_path=None):
    """
    Get all valid insertions from the data base
    """
    insertions = None
    if local_file_path is None:
        local_file_path = str(utils.EXAMPLES_DATA_FOLDER.joinpath('valid_insertions.p'))
    if not force_query and os.path.exists(os.path.abspath(local_file_path)):
            insertions = pickle.load(open(local_file_path, 'rb'))
            print('Using local file for insertions', local_file_path)
    else:
            dq = 'session__project__name__icontains,ibl_neuropixel_brainwide_01,session__qc__lt,50,'
            dq += '~json__qc,CRITICAL,session__extended_qc__behavior,1,json__extended_qc__tracing_exists,True,'
            dq += '~session__extended_qc___task_stimOn_goCue_delays__lt,0.9,'
            dq += '~session__extended_qc___task_response_feedback_delays__lt,0.9,'
            dq += '~session__extended_qc___task_wheel_move_before_feedback__lt,0.9,'
            dq += '~session__extended_qc___task_wheel_freeze_during_quiescence__lt,0.9,'
            dq += '~session__extended_qc___task_error_trial_event_sequence__lt,0.9,'
            dq += '~session__extended_qc___task_correct_trial_event_sequence__lt,0.9,'
            dq += '~session__extended_qc___task_reward_volumes__lt,0.9,'
            dq += '~session__extended_qc___task_reward_volume_set__lt,0.9,'
            dq += '~session__extended_qc___task_stimulus_move_before_goCue__lt,0.9,'
            dq += '~session__extended_qc___task_audio_pre_trial__lt,0.9'
            query_result = one.alyx.rest('insertions', 'list', django=dq)
            #insertions = one.alyx.rest('insertions', 'list', django=query)
            insertions = []
            for insertion in query_result:
                    insertions.append(insertion)
            pickle.dump(insertions, open(local_file_path, 'wb'))
            #points_query = one.alyx.rest('')
    return insertions


def get_insertions_data(one=None, ba=None, insertions=None, start=None, end=None, 
                        exceptions=None, force_query=False, local_file_path=None):
    """
    Get all valid insertions and related data from the DB.

    """
    insertions_data = None
    if local_file_path is None:
        local_file_path = str(utils.EXAMPLES_DATA_FOLDER.joinpath('valid_insertions_data.p'))
    if not force_query and os.path.exists(os.path.abspath(local_file_path)):
            insertions_data = pickle.load(open(local_file_path, 'rb'))
            print('Using local file for insertions data', local_file_path)
    else:
        if one is None:
            one = ONE()
        if ba is None:
            ba = atlas.AllenAtlas()
        if insertions is None:
            insertions = get_valid_insertions()
        num_insertions = len(insertions)

        print('')
        print('')
        print('Got', len(insertions), 'valid insertions (that do not necessarily have data)')

        # Retrieving XYZ point neuron coordinates goes like this: 
        # spikes.clusters -> clusters.channels -> channels.xyz 
        # but in many cases, channels.xyz (a file called mlapdv.npy) does not exist
        # so you have to compute XYZ coordinates given spikes depths and insertion XYZ
        dtypes = [
            'spikes.depths',
            'spikes.amps',
            'spikes.times',
            'spikes.clusters',
            'clusters.channels',
            'clusters.mlapdv'
        ]

        # Filter insertions and retain only valid data because even 
        # the insertions deemed valid are sometimes not what we want
        insertions_data = {}

        if not isinstance(start, int):
            start = 0
        if not isinstance(end, int):
            end = num_insertions
        start = max(0, start)
        end = min(num_insertions, end)
        print('Getting data for insertions', start, 'to', end, 'out of', num_insertions)
        for loop_id in range(start, end):
            insertion = insertions[loop_id]
            probe_id = insertion['id']
            if exceptions is not None and probe_id in exceptions:
                print('Probe', loop_id, 'with id', probe_id, 'is in exclusion list (probably because of invalid data)')
                continue
            data_sets = one.alyx.rest('datasets', 'list', probe_insertion=probe_id, django=f'dataset_type__name__in,{dtypes}')
            if data_sets is None or len(data_sets) < 1:
                print('Probe', loop_id, ': no dataset found (!) for id', probe_id)
                continue
            print('Probe', loop_id, 'using', len(data_sets), 'data sets for id', probe_id)
            insertions_data[probe_id] = data_sets
        
        print(len(insertions_data), 'valid insertions found on', num_insertions)
        print('')
        pickle.dump(insertions_data, open(local_file_path, 'wb'))
    return insertions_data


def get_point_neurons_data(one=None, ba=None, insertions=None, start=0, end=None, 
                            recompute_positions=True, probe_ids=None, 
                            exceptions=['825ba9b8-ce03-49b7-b1a8-4d85ae2185af'], 
                            local_file_path=None):
    """
    Get all point neurons data sets that IBL has measured/recorded.
    Warning, this can result in veeery long download time initially.
    I had to download 85GB of data (including relevant spike data) in May 2021. 
    And the best part is that the position of the neurons in the end
    is put in a pickle file less than 1MB. :)
    :param one: ONE connection object
    :param ba: BrainAtlas object
    :param insertions: 
    :param start: 
    :param end: 
    :param recompute_positions: 
    :param probe_ids: 
    :param exceptions: 
    :param local_file_path: 
    """
    if local_file_path is None:
        local_file_path = str(utils.EXAMPLES_DATA_FOLDER.joinpath('ibl_point_neurons.npz'))
    
    use_local_data = not recompute_positions and os.path.exists(os.path.abspath(local_file_path))
    if use_local_data:
        result = np.load(local_file_path, allow_pickle=True)
        print('Point neurons data loaded (using local npz storage)', local_file_path)
        insertion_ids = result['insertion_ids']
        xyz_positions = result['xyz_positions']
        xyz_resolved_mask = result['xyz_resolved_mask']
        data_mask = result['data_mask']
        return insertion_ids, xyz_positions, xyz_resolved_mask, data_mask

    if one is None:
        one = ONE()
    if ba is None:
        ba = atlas.AllenAtlas()
    if insertions is None:
        insertions = get_valid_insertions()
    num_insertions = len(insertions)
    insertions_data = get_insertions_data(one, ba, insertions, start, end, exceptions)

    insertion_ids = [''] * num_insertions
    xyz_positions = [[]] * num_insertions
    xyz_resolved_mask = [False] * num_insertions
    data_mask = [False] * num_insertions

    loop_id = -1
    got_ids_filter = isinstance(probe_ids, dict) or isinstance(probe_ids, list)
    
    # Lots of try catch ahead! This is to handle all the cases where things might be missing or failing
    for probe_id in insertions_data:
        loop_id += 1
        data_sets = insertions_data[probe_id]
        if got_ids_filter and probe_id not in probe_ids:
            continue
        try:
            local_files = one.download_datasets(data_sets)
        except Exception as e:
            print('Error downloading dataset', e)
            continue
        
        if local_files is None:
            print('Local files not found for', probe_id)
            continue
        '''
        TODO: this is where you start using spike data.
        try:
            spikes = alf.io.load_object(local_files[0].parent, 'spikes')
        except AttributeError:
            continue
        '''
        try:
            clusters = alf.io.load_object(local_files[0].parent, 'clusters')
        except AttributeError:
            continue
        #channels = clusters['channels']

        print('Processing insertion', loop_id+1, 'on', len(insertions_data), '- Probe id', probe_id)
        # When mlapdv dataset does not exist, it means we have to 
        # estimate xyz with the given spike depths and insertion xyz
        if 'mlapdv' in clusters:
            xyz_positions[loop_id] = clusters['mlapdv']
        else:
            '''
            Exemple of an insertion that throws a KeyError:
            0393f34c-a2bd-4c01-99c9-f6b4ec6e786d
            KeyError: ibl.brainbox.io '2021-04-20T15:49:05_anup.khanal'
            So we have an id but something went wrong in the data pipeline and it got invalid
            '''
            insertion = one.alyx.rest('insertions', 'read', id=probe_id)
            #print('Is insertion equal to itself???', insertion==insertion2, insertion['id'], insertion2['id'])
            try:
                channel_data = load_channels_from_insertion(insertion, one=one, ba=ba)
            except Exception as e:
                print(e)
                continue
            # Units are in meters so we make that consistent with mlapdv data sets that are in microns
            xyz_positions[loop_id] = channel_data * 1000000

        xyz_resolved_mask[loop_id] = 'mlapdv' in clusters
        data_mask[loop_id] = True
        insertion_ids[loop_id] = probe_id
        
    #if not isinstance(xyz_positions, np.ndarray):
    xyz_positions = np.array(xyz_positions)
    if local_file_path is not None:
        print('Saving processed data under', local_file_path)
        np.savez_compressed(local_file_path, insertion_ids=insertion_ids, xyz_positions=xyz_positions, 
        xyz_resolved_mask=xyz_resolved_mask, data_mask=data_mask)
    
    print('Point neurons data downloaded')
    return insertion_ids, xyz_positions, xyz_resolved_mask, data_mask


def main():
    # Test probe ids. The first one has MLAPDV data, the others have their channels computed instead.
    probe_ids = ['00a824c0-e060-495f-9ebc-79c82fef4c67',
                '0143d3fe-79c2-4922-8332-62c3e4e0ba85',
                '4762e8ed-4d94-4fd7-9522-e927f5ffca74',
                '4755877d-fd14-42b3-bc15-e1996d153015']

    # If you want to run the test with only the selection from above, then comment the below line
    probe_ids = None

    '''
    There is an error for this data:
    Spike data not found: 56f2a378-78d2-4132-b3c8-8c1ba82be598
    '''
    data = get_point_neurons_data(probe_ids=probe_ids, recompute_positions=False)
    insertion_ids, xyz_positions, xyz_resolved_mask, data_mask = data
    positions = []
    values = []
    loop_id = 0
    # xyz_positions is a sparse array so we take only valid data from it
    insertions = xyz_positions[data_mask]

    for channels in insertions:
        channels_positions = channels if isinstance(channels, list) else channels.tolist()
        #inside_brain_positions = get_points_inside_brain(viewer, channels_positions)
        positions += channels_positions
        mock_ids = np.ones(len(channels_positions))*loop_id
        values += mock_ids.tolist()
        loop_id += 1
    positions = np.array(positions)
    values = np.array(values)
    
    av = np.copy(values) * np.random.random(len(values))
    bv = np.copy(values) * np.random.random(len(values))
    cv = np.copy(values) * np.random.random(len(values))
    values = np.c_[values, av, bv, cv]

    viewer = MouseBrainViewer()
    print('Data report: using', len(insertions), 'insertions and', len(positions), 'channels (point neurons)')
    print('-> Bounds:', np.amin(positions, axis=0), np.amax(positions, axis=0))
    # This starts the viewer with two volumes: the segmented atlas and the DWI of the mouse brain
    viewer.initialize(resolution=50, add_dwi=True, dwi_color_map='viridis', 
                        add_atlas=True, mapping='Allen', dark_mode=True, embed_ui=True)

    # Below are a few optimizations you could use if your computer 
    # is too slow to visualize the result:

    # 1. We add all the points together as one object. 
    # The fewer objects on your 3D scene, the faster it will render.

    # 2. Activate screen space mode so that the points are in 2D.
    # This can literally change the game if you have a low-end computer

    # 3. Do not trim outliers, because the computation to know if each
    # and every point is within the brain is quite demanding.
    points = viewer.add_points(positions, radius=8, values=values, screen_space=True, 
                                noise_amount=100, trim_outliers=False, add_to_scene=True)

    # You will need some CPU/GPU power to handle volume 
    # transparency and all those semi-transparent points
    points.alpha(0.5)

    # Finally, before showing everything, we select the latest added object (the points)
    viewer.select(-1)
    viewer.show()


if __name__ == '__main__':
    main()


import argparse
import numpy as np
from iblviewer.qt_application import ViewerApp
from iblviewer.application import Viewer
got_ibllib = True
from iblviewer.mouse_brain import MouseBrainViewer



"""
Project: IBL Viewer

Description: this is a fast and interactive 3D viewer for exploring
and analysing volumes, surfaces, points and lines.
It's based on Python, VTK and partly on vedo (a VTK wrapper).

In the context of the International Brain Laboratory, this viewer is
used by neuroscientists to perform analysis of data models in the
context of the Allen Mouse CCF v3 atlas. 

Copyright: 2021 Nicolas Antille, International Brain Laboratory
License: MIT
"""
# From stackoverflow
def str2bool(v):
    if isinstance(v, bool):
        return v
    elif v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

class IBLViewer():

    def __init__(self):
        self.viewer = None
        self.qt_app = None
        self.args = None

    def parse_args(self, parser=None):
        """
        Standard argument parser for iblviewer. Make sure you do not use the below argument shortcuts:
        -s, -t, -ui, -r, -m, -d, -atlas, -dwi, -cm, -nc, -na
        :param parser: An existing ArgumentParser that will be updated with standard arguments.
            If None, a new one is created (default).
        :return: ArgumentParser
        """
        if parser is None:
            parser = argparse.ArgumentParser(description='International Brain Viewer based on VTK')

        parser.add_argument('-neuro', dest='neuroscience', type=str2bool, default=True, 
        help='Whether the viewer starts in Neuroscience mode with Mouse Brain volume preset (1) or as generic 3D viewer (0)')

        parser.add_argument('-t', dest='test_data', type=str2bool, default=False, 
        help='Whether a simple random set of points is added as test data')

        parser.add_argument('-ui', dest='ui', type=int, default=1, 
        help='User interface. 0 for none, 1 for Qt, 2 for embed')

        # More command-line options are given in the context of neuroscience
        parser.add_argument('-r', dest='resolution', type=int, default=50, 
        help='Volume resolution. Possible values: 100, 50, 25, and 10. Units are in microns.\
            The 10um volume takes a lot of RAM (and some time to load)')
        
        parser.add_argument('-m', dest='mapping', type=str, default='Allen', 
        help='Volume mapping name. Either Allen (default value) or Beryl (IBL specific simplified mapping).')
        
        parser.add_argument('-d', dest='dark_mode', type=str2bool, default=True, 
        help='Enable (1) or disable (0) dark mode.')
        
        parser.add_argument('-atlas', dest='add_atlas', type=str2bool, default=True, 
        help='If the Allen Atlas volume should be added to the viewer')
        
        parser.add_argument('-dwi', dest='add_dwi', type=str2bool, default=False, 
        help='If the Allen Atlas raw DWI volume should be added to the viewer')
        
        parser.add_argument('-cm', dest='color_map', type=str, default='viridis', 
        help='Color map for custom data mapped on to the Allen atlas volume')
        
        parser.add_argument('-nc', dest='nan_color', type=float, default=0.65, 
        help='Gray color (between 0 and 1) for regions that have no assigned value')
        
        parser.add_argument('-na', dest='nan_alpha', type=float, default=0.5, 
        help='Alpha (opacity) value for regions that have no assigned value')

        args = parser.parse_args()
        self.args = args
        return args

    def launch(self, callable=None, stats_callable=None, args=None, 
                jupyter=False, neuroscience=True, **kwargs):
        """
        Start the 3D viewer according to parameters given in the console
        :param callable: Function that will be called when the 3D viewer is initialized
        :param stats_callable: Function that will be called when statistics are updated,
            when a selection or sub selection changes in the 3D viewer. Available when 
            you use the Qt UI only.
        :param args: Any existing ArgumentParser. If None, a new IBL standard one is created.
        :param jupyter: Whether you launch the viewer within a jupyter notebook or lab
        :param neuroscience: Whether the viewer in jupyter is started in neuroscience mode or not
        :param kwargs: All further keyword arguments set to viewer.initialize() method (for jupyter mode)
        :return: Either a qt_application.ViewerApp (if Qt) or a viewer instance (mouse_brain.MouseBrainViewer
            or application.)
        """
        ibllib_msg = 'The viewer is set to start in neuroscience mode but you do not have ibllib '
        ibllib_msg += 'optional module installed.\n\nPlease run pip install ibllib and run the viewer '
        ibllib_msg += 'again if you want to start in neuroscience mode.\n\n'
        ibllib_msg += 'Alternatively, you may use the viewer in standard mode with random points for test: iblviewer -neuro 0 -t 1\n'
        if jupyter:
            if neuroscience and not got_ibllib:
                print(ibllib_msg)
                exit()
            if neuroscience:
                # This a computational neuroscience environment, in this case focused
                # on the Allen Brain Atlas and International Brain Laboratory data models
                self.viewer = MouseBrainViewer()
            else:
                # This is a generic 3D viewer, not related to neuroscience
                self.viewer = Viewer()
            self.viewer.initialize(**kwargs)
            return self.viewer.show()

        if args is None:
            args = self.args
            if args is None:
                args = self.parse_args()
                self.args = args

        if args.neuroscience:
            # This a computational neuroscience environment, in this case focused
            # on the Allen Brain Atlas and International Brain Laboratory data models
            self.viewer = MouseBrainViewer()
        else:
            # This is a generic 3D viewer, not related to neuroscience
            self.viewer = Viewer()

        qt_mode = args.ui == 1
        if qt_mode:
            self.qt_app = ViewerApp()
            if args.neuroscience:
                # viewer.initialize(...) method will be called internally with the expanded args
                self.qt_app.initialize(viewer=self.viewer, callable=callable, stats_callable=stats_callable, 
                                embed_ui=args.ui==2, offscreen=False, dark_mode=args.dark_mode, 
                                resolution=args.resolution, mapping=args.mapping, add_dwi=args.add_dwi, 
                                add_atlas=args.add_atlas)
            else:
                new_function = callable
                if args.test_data:
                    # Test data
                    points = np.random.random((500, 3)) * 1000
                    def new_function(viewer):
                        if callable is not None:
                            callable(viewer)
                        viewer.add_spheres(points, radius=10)
                # viewer.initialize(...) method will be called internally with the expanded args
                self.qt_app.initialize(viewer=self.viewer, callable=new_function, stats_callable=stats_callable,
                                embed_ui=args.ui==2, offscreen=False, dark_mode=args.dark_mode)
            return self.qt_app
            # Any code below here is only executed once you quit the Qt application

        else:
            if args.neuroscience:
                self.viewer.initialize(resolution=args.resolution, mapping=args.mapping, add_dwi=args.add_dwi,
                                    add_atlas=args.add_atlas, embed_ui=args.ui==2, offscreen=False, 
                                    jupyter=jupyter, dark_mode=args.dark_mode)
            else:
                self.viewer.initialize(embed_ui=args.ui==2, offscreen=qt_mode, jupyter=jupyter)
                if args.test_data:
                    # Test data
                    points = np.random.random((500, 3)) * 1000
                    self.viewer.add_points(points, radius=10)

            if callable is not None:
                callable(self.viewer)
            self.viewer.show()
            return self.viewer

def main(auto_close_viewer=True):
    app = IBLViewer()
    app.launch()
    if auto_close_viewer:
        app.viewer.close()
    return app

if __name__ == '__main__':
    app = main()


from dataclasses import dataclass
import os
import logging
import numpy as np
import requests
from pathlib import Path
import textwrap

import vedo
import nrrd
import ibllib.atlas
from ibllib.atlas.regions import BrainRegions
from iblutil.numerical import ismember

from iblviewer.application import Viewer
from iblviewer.volume import VolumeController, VolumeModel, LUTModel
import iblviewer.utils as utils

ALLEN_ATLAS_RESOLUTIONS = [10, 25, 50, 100]
ALLEN_ATLASES = {'base_url': 'http://download.alleninstitute.org/informatics-archive',
                'mouse_ccf_folder': '/current-release/mouse_ccf',
                'atlas_folder': '/annotation/ccf_2017', 'atlas_prefix': '/annotation_',
                'dwi_folder': '/average_template', 'dwi_prefix': '/average_template_',
                'volume_extension': '.nrrd'}

# Default origin is "bregma", an origin defined at the center of the XY axes (not on Z)
# For reference, bregma is np.array([5739.0, 5400.0, 332.0])
# And for some reason, it's not exactly in the middle of X axis (should be 5400.0)...
IBL_BREGMA_ORIGIN = ibllib.atlas.ALLEN_CCF_LANDMARKS_MLAPDV_UM['bregma']
BASE_PATH = utils.split_path(os.path.realpath(__file__))[0]
REMAPPED_VOLUME_SUFFIX = '_remapped'
_logger = logging.getLogger('ibllib')
LUT_VERSION = 'v01' # version 01 is the lateralized version


class AllenAtlasExt(ibllib.atlas.AllenAtlas):
    """
    This overwrites the constructor of AllenAtlas that is not designed to be used for the
    public, that is people outside of IBL. Typically, you'd want to display the Allen volume
    data in this viewer and perform additional tasks (such as loading your own extra data)
    with other libraries. Dev note: I'm forced to copy and modify the whole constructor in this case.

    Instantiates an atlas.BrainAtlas corresponding to the Allen CCF at the given resolution
    using the IBL Bregma and coordinate system
    """
    @staticmethod
    def _read_volume(file_volume):
        if file_volume.suffix == '.nrrd':
            volume, _ = nrrd.read(file_volume, index_order='C')  # ml, dv, ap
            # we want the coronal slice to be the most contiguous
            volume = np.transpose(volume, (2, 0, 1))  # image[iap, iml, idv]
        elif file_volume.suffix == '.npz':
            volume = np.load(str(file_volume), allow_pickle=True)['arr_0']
        return volume

    def __init__(self, res_um=25, brainmap='Allen', scaling=np.array([1, 1, 1]),
                image_file_path=None, label_file_path=None):
        """
        :param res_um: 10, 25 or 50 um
        :param brainmap: defaults to 'Allen', see ibllib.atlas.BrainRegion for re-mappings
        :param scaling: scale factor along ml, ap, dv for squeeze and stretch ([1, 1, 1])
        :return: atlas.AllenAtlasGP
        """
        xyz2dims = np.array([1, 0, 2]) # this is the c-contiguous ordering
        dims2xyz = np.array([1, 0, 2])
        # we use Bregma as the origin
        self.res_um = res_um
        dxyz = self.res_um * 1e-6 * np.array([1, -1, -1]) * scaling

        if label_file_path is None:
            # No point in going further
            return

        regions = BrainRegions()
        ibregma = (ibllib.atlas.ALLEN_CCF_LANDMARKS_MLAPDV_UM['bregma'] / self.res_um)
        #image_path = atlas_path.joinpath(f'average_template_{res_um}.nrrd')
        # It is really unfortunate that users are forced to have in memory two volumes: the atlas and the DWI!
        image = AllenAtlasExt._read_volume(Path(image_file_path))
        label = self.remap_atlas(label_file_path, regions, ibregma)

        # This calls BrainAtlas, the mother class of AllenAtlas because we want to overwrite it
        super(ibllib.atlas.AllenAtlas, self).__init__(image, label, dxyz, regions,
        ibregma, dims2xyz=dims2xyz, xyz2dims=xyz2dims)

    def remap_atlas(self, local_file_path, regions=None, ibregma=None):
        """
        Remap the atlas label into a usable volume
        """
        file_path = Path(local_file_path)
        parent_folder = file_path.parent.absolute()
        npz_name = f'{file_path.stem}_lut_{LUT_VERSION}.npz'
        remapped_file_path = parent_folder.joinpath(npz_name)
        if not remapped_file_path.exists():
            volume = self._read_volume(file_path)
            if regions is None:
                # Encapsulating the lateralization, that should be a given
                volume = self.lateralize(volume, regions, ibregma)
            _logger.info(f"saving {remapped_file_path} ...")
            np.savez_compressed(str(remapped_file_path), volume)
        else:
            volume = np.load(str(remapped_file_path))['arr_0']
        return volume

    def lateralize(self, label, regions=None, ibregma=None):
        """
        Breaks the symmetry in regions labels in the Allen Mouse Atlas where the id
        of a region in the left hemisphere is the same as the same region in the right
        hemisphere. But if we want to map recordings to the brain, we need separate ids.
        :param label: Segmented volume
        :return: Modified volume
        """
        _logger.info("computing brain atlas annotations lookup table")
        if regions is None:
            regions = BrainRegions()
        if ibregma is None:
            ibregma = (ibllib.atlas.ALLEN_CCF_LANDMARKS_MLAPDV_UM['bregma'] / self.res_um)
        xyz2dims = np.array([1, 0, 2])  # this is the c-contiguous ordering
        # lateralize atlas: for this the regions of the left hemisphere have primary
        # keys opposite to to the normal ones
        lateral = np.zeros(label.shape[xyz2dims[0]])
        lateral[int(np.floor(ibregma[0]))] = 1
        lateral = np.sign(np.cumsum(lateral)[np.newaxis, :, np.newaxis] - 0.5)
        label = label * lateral
        _, im = ismember(label, regions.id)
        label = np.reshape(im.astype(np.uint16), label.shape)
        return label



@dataclass
class IBLAtlasModel():
    """
    IBL Atlas is a wrapper for the Allen Atlas with added features.
    The volume is also modified such that it fits functional needs.
    """
    origin: np.ndarray = IBL_BREGMA_ORIGIN
    atlas: ibllib.atlas.AllenAtlas = None
    atlas_lut: LUTModel = None
    atlas_mapping_ids: list = None
    ibl_back_end: bool = True

    IBL_TRANSPOSE = (1, 0, 2)

    atlas_volume: VolumeModel = None
    dwi_volume: VolumeModel = None

    def initialize(self, resolution=None, ibl_back_end=True):
        """
        Get Allen Atlas metadata (.csv file that is copied into ibllib.atlas) and volume files
        :param resolution: Volume resolution in microns
        """
        if resolution not in ALLEN_ATLAS_RESOLUTIONS:
            resolution = ALLEN_ATLAS_RESOLUTIONS[-1]
        if resolution == ALLEN_ATLAS_RESOLUTIONS[0]:
            print('Warning: the Allen Atlas at 10um resolution will require over 10GB of RAM')
        self.resolution = resolution
        self.use_ibl_back_end = ibl_back_end

        try:
            self.atlas = ibllib.atlas.AllenAtlas(resolution)
        except Exception as e:
            atlas_volume_url = self.get_allen_volume_url(resolution)
            atlas_volume_name = self.get_allen_volume_file_name(resolution)
            atlas_file = requests.get(atlas_volume_url)
            atlas_volume_path = './' + atlas_volume_name
            open(atlas_volume_path, 'wb').write(atlas_file.content)

            dwi_volume_url = self.get_allen_volume_url(resolution, True)
            dwi_volume_name = self.get_allen_volume_file_name(resolution, True)
            dwi_file = requests.get(dwi_volume_url)
            dwi_volume_path = './' + dwi_volume_name
            open(dwi_volume_path, 'wb').write(dwi_file.content)
            self.atlas = AllenAtlasExt(resolution, image_file_path=dwi_volume_path, label_file_path=atlas_volume_path)

        self.atlas_mapping_ids = list(self.atlas.regions.mappings.keys())
        # Further than initialize(), you will need to either call get_atlas_model() or get_dwi_model()

    def get_atlas_model(self, atlas_mapping=None):
        """
        Get a VolumeModel instance that represents the atlas (segmented) volume
        :param atlas_mapping: IBL Mapping that we want to use on this atlas.
            See ibllib.atlas.AllenAtlas.regions.mappings.keys()
        :return: VolumeModel
        """
        if self.atlas is None:
            print('No atlas!')
            return
        if self.atlas_volume is not None:
            return self.atlas_volume
        self.atlas_volume = VolumeModel(resolution=self.resolution, base_color_map=self.atlas.regions.rgb)
        self.atlas_volume.data = self.get_mapped_volume(self.atlas.label, atlas_mapping)
        self.atlas_volume.mapping = atlas_mapping
        self.atlas_volume.data_type = VolumeModel.SEGMENTED
        self.atlas_volume.name = 'Allen Mouse CCF v3 Atlas volume'
        # IBL convention
        self.atlas_volume.lateralized = atlas_mapping is not None and '-lr' in atlas_mapping
        if self.ibl_back_end:
            # The IBL back-end uses a different convention for memory representation
            # so we are forced to untranspose the volume...
            self.atlas_volume.transpose(IBLAtlasModel.IBL_TRANSPOSE)
        self.atlas_volume.build_lut(color_map=self.atlas.regions.rgb, make_active=True)
        self.atlas_volume.compute_size()
        return self.atlas_volume

    def get_dwi_model(self):
        """
        Get a VolumeModel instance of the DWI volume image
        :return: VolumeModel
        """
        if self.atlas is None:
            return
        if self.dwi_volume is not None:
            return self.dwi_volume
        self.dwi_volume = VolumeModel(resolution=self.resolution)
        self.dwi_volume.data = self.atlas.image
        self.dwi_volume.data_type = VolumeModel.RAW
        self.dwi_volume.name = 'Allen Mouse CCF v3 DWI volume'
        if self.ibl_back_end:
            self.dwi_volume.transpose(IBLAtlasModel.IBL_TRANSPOSE)
        self.dwi_volume.compute_size()
        return self.dwi_volume

    def get_num_regions(self):
        """
        Get how many regions are labelled
        """
        return self.atlas.regions.id.size

    def get_value_from_scalar_map(self, scalar):
        """
        Reverse look-up in array to find a corresponding value
        :param scalar: Scalar value
        :return: Raw volume value
        """
        scalar_map = self.atlas_volume.luts.current.scalar_map
        if scalar_map is None:
            return
        for value in range(len(scalar_map)):
            if scalar_map[value] is None:
                continue
            #print(scalar - scalar_map[value])
            if scalar_map[value] == scalar:
                return value

    def get_mapped_data(self, value):
        """
        Given a value from the segmented volume, we retrieve useful info
        :param value: Value from the volume
        :return: Dictionary of corresponding data
        """
        if value is None:
            return
        value = int(value)
        region_id = self.atlas.regions.id[value]
        region_data = self.atlas.regions.get(region_id)

        data_lut = self.atlas_volume.luts.current
        region_name = region_data.name[0].title()
        color = data_lut.color_map[value][1]
        alpha = 1.0
        if data_lut.alpha_map is not None:
            alpha = data_lut.alpha_map[value][1]
        scalar = None
        scalar_map = data_lut.scalar_map
        if scalar_map is not None:
            if isinstance(scalar_map, dict):
                scalar = scalar_map.get(value)
            else:
                scalar = scalar_map[value]
        return {'scalar':scalar, 'region_id':region_id, 'color':color, 'alpha':alpha,
                'region_name':region_name, 'region_data':region_data}

    def remap(self, ids, source='Allen', dest='Beryl'):
        """
        Remap ids/scalar values from source to destination
        Function by Olivier Winter
        :param ids: List of ids
        :param source: Source mapping
        :param dest: Destination mapping
        """
        #from ibllib.atlas import BrainRegions as br
        _, inds = ismember(ids, self.atlas.regions.mappings[source])
        return self.atlas.regions.mappings[dest][inds]

    def get_allen_volume_file_name(self, resolution, raw_image=False):
        """
        Get the Allen volume file name given its resolution
        :param resolution: Resolution of the volume
        :param raw_image: Whether we want the raw volume file name or the segmented one
        :return: String
        """
        file_name = ALLEN_ATLASES['dwi_prefix'] if raw_image else ALLEN_ATLASES['atlas_prefix']
        file_name += str(resolution) + ALLEN_ATLASES['volume_extension']
        return file_name

    def get_allen_volume_url(self, resolution, raw_image=False):
        """
        Construct a URL with which we can download data sets
        :param resolution: Volume resolution, either 10, 25, 50 or 100 (um)
        :param raw_image: Whether the volume is the segmented one (aka the atlas) or the DWI
        :return: String
        """
        url = ALLEN_ATLASES['base_url'] + ALLEN_ATLASES['mouse_ccf_folder']
        url += ALLEN_ATLASES['dwi_folder'] if raw_image else ALLEN_ATLASES['atlas_folder']
        url += self.get_allen_volume_file_name(resolution, raw_image)
        return url

    def load_volume(self, file_path, remap_scalars=False, mapping=None, make_current=True):
        """
        Load a volume data file. Supports NRRD and many other formats thanks to vedo/VTK
        :param file_path: Volume file path. Could support other file types easily.
        :param remap_scalars: Whether scalar values in the volume are replaced by
            their row id from a mapping that stores. This is necessary in the case of segmented
            volumes with regions that have a discontinuous id.
        :param mapping: Pandas Series or a Dictionary
        :param make_current: Set the volume data as the current one
        :return: 3D array
        """
        return super().load_volume(file_path, remap_scalars, self.atlas.regions.id, make_current)

    def get_name(self, *args):
        """
        Get full name for a model, separated by underscores
        :return: String
        """
        return '_'.join(args)

    def get_mapped_volume(self, volume, atlas_mapping=None, ibl_back_end=True):
        """
        Set the volume data according to a mapping
        :param volume: Given volume to display
        :param atlas_mapping: Mapping, either a string for the name of the mapping or an integer.
        :param ibl_back_end: If you are not using ibllib and want to load your own volume, set this to False
            so that there will be no transposition of the volume (needed for the ones from IBL)
        :return: Volume nd array
        """
        volume_data = None
        if ibl_back_end:
            if isinstance(atlas_mapping, int):
                if atlas_mapping > len(self.atlas_mapping_ids) - 1:
                    #logging.error('[AtlasModel.get_mapped_volume()] could not find atlas mapping with id ' + str(atlas_mapping) + '. Returning raw volume...')
                    return volume
                map_id = self.atlas_mapping_ids[atlas_mapping]
            elif atlas_mapping is None:
                map_id = self.atlas_mapping_ids[0]
            else:
                map_id = atlas_mapping

            # This mapping actually changes the order of the axes in the volume...
            volume_data = self.atlas.regions.mappings[map_id][volume]
            #logging.info('Loaded atlas volume with ' + map_id + ' mapping')
        else:
            volume_data = volume
        return volume_data

    def get_region_and_row_id(self, acronym):
        """
        Get region and row id given an acronym
        :param acronym: Acronym of a brain region
        :return: Region id and row id
        """
        ind = np.where(self.atlas.regions.acronym == acronym)[0]
        if ind.size < 1:
            return None, None
        return self.atlas.regions.id[ind], ind

    def get_regions_mask(self, region_ids, alpha_map=None):
        """
        Build an alpha map that reveals only the given region ids
        :param region_ids: List or numpy array of region ids
        :param alpha_map: Optional alpha map that will be modified. If None provided, the method will attempt
            to use the current active alpha map
        :return: 2D numpy array with region ids and corresponding alpha values
        """
        if alpha_map is None:
            alpha_map = self.lut.alpha_map
        if alpha_map is None:
            #logging.error('[Method build_regions_alpha_map()] requires that an alpha map is created by build_regions_alpha_map()')
            return
        new_alpha_map = np.zeros_like(alpha_map).astype(float)
        new_alpha_map[:, 0] = alpha_map[:, 0]
        new_alpha_map[region_ids, 1] = alpha_map[region_ids, 1]
        self.lut.sec_alpha_map = new_alpha_map
        return new_alpha_map


class MouseBrainViewer(Viewer):
    """
    This is your entry point to International Brain Laboratory data visualization
    """

    def __init__(self):
        """
        Constructor
        """
        super().__init__()
        self.bounding_mesh = None
        self.atlas_controller = None
        self.dwi_controller = None
        self.ibl_model = None
        # Shortcut for users
        self.ibl_transpose = IBLAtlasModel.IBL_TRANSPOSE

    def initialize(self, resolution=25, mapping='Beryl', add_atlas=True, add_dwi=False,
                    dwi_color_map='viridis', dwi_alpha_map=None, local_allen_volumes_path=None,
                    offscreen=False, jupyter=False, embed_ui=False, embed_font_size=15,
                    plot=None, plot_window_id=0, num_windows=1, render=False, dark_mode=False):
        """
        Initialize the controller, main entry point to the viewer
        :param resolution: Resolution of the atlas volume.
            Possible values are 10 (requires a lot of RAM), 25, 50, 100. Units are in microns
        :param mapping: Optional mapping value. In the context of IBL, there is 'Allen' for the standard Allen map
            and 'Beryl' (random name) which aggregates cortical layers as one.
        :param add_atlas: Whether the Atlas is included in the viewer
        :param add_dwi: Whether the diffusion weighted imaging is loaded in the viewer (same boundaries as atlas)
        :param context: Context of the visualization
        :param embed_ui: Whether the UI is embed within the VTK window
        :param embed_font_size: Embed font size. Defaults to 16 points. You might need larger values
            in case you have a small screen with high dpi (but VTK methods fail to detect that).
        :param jupyter: Whether we're running from a jupyter notebook or not
        :param plot: A vedo Plotter instance. You can either create it by yourself before hand, in case you want
            to have multiple windows with other stats or let the controller create a new one
        :param plot_window_id: Sub-window id where the 3D visualization will be displayed
        :param num_windows: Number of subwindows, in case you want to display your own stuff later
        :param dark_mode: Whether the viewer is in dark mode
        """
        self.model.title = 'IBL Viewer'

        # ibllib works with two volumes at the same time: the segmented volume (called 'label')
        # and the DWI volume (called 'image')
        self.ibl_model = IBLAtlasModel()
        self.ibl_model.initialize(resolution, local_allen_volumes_path)
        self.model.origin = self.ibl_model.origin

        # Neuroscience specific
        self.model.probe_initial_point1 = self.model.origin
        self.model.probe_initial_point2 = self.model.origin + [0, 0, 8000]

        super().initialize(offscreen, jupyter, embed_ui, embed_font_size,
                            plot, plot_window_id, num_windows, dark_mode)

        # A VolumeController has a unique volume as target so if we want to visualize both volumes, we create two views
        if add_atlas:
            self.ibl_model.get_atlas_model(mapping)
            self.add_atlas_segmentation()
        if add_dwi:
            self.ibl_model.get_dwi_model()
            if dwi_alpha_map is None:
                #dwi_alpha_map = np.ones(516)
                #dwi_alpha_map[:140] = 0
                dwi_alpha_map = [0.0, 1.0, 1.0]
            self.add_atlas_dwi(dwi_color_map, dwi_alpha_map)

        try:
            self.load_bounding_mesh()
        except Exception:
            pass

        #light = vedo.Light(self.model.IBL_BREGMA_ORIGIN - [0, 0, 1000], c='w', intensity=0.2)
        #self.plot.add(light)

        # By default, the atlas volume is our target
        if self.atlas_controller is not None:
            self.select(self.atlas_controller.actor)
        elif self.dwi_controller is not None:
            self.select(self.dwi_controller.actor)

        if self.model.selection is not None:
            #self.add_origin()
            self.handle_lut_update()
            self.set_left_view()

    def add_atlas_segmentation(self):
        """
        Add the Allen Atlas segmented volume (aka label)
        """
        if isinstance(self.atlas_controller, VolumeController):
            return
        self.atlas_controller = VolumeController(self.plot, self.ibl_model.atlas_volume,
                                                alpha_unit_upper_offset=0.1, center_on_edges=True)
        self.register_controller(self.atlas_controller, self.atlas_controller.get_related_actors())

    def add_atlas_dwi(self, color_map, alpha_map):
        """
        Add the Allen Atlas diffusion weighted image
        :param color_map: Color map for the volume
        :param alpha_map: Alpha map for the volume
        """
        if isinstance(self.dwi_controller, VolumeController):
            return
        self.dwi_controller = VolumeController(self.plot, self.ibl_model.dwi_volume,
                                                center_on_edges=True)
        self.dwi_controller.set_color_map(color_map, alpha_map)
        self.register_controller(self.dwi_controller, self.dwi_controller.get_related_actors())

    def load_bounding_mesh(self, add_to_scene=False, alpha_on_scene=0.3):
        """
        Load the bounding mesh of the mouse brain that represents its approximate pial limit
        """
        self.bounding_mesh = utils.load_surface_mesh('997')
        if add_to_scene:
            self.bounding_mesh.alpha(alpha_on_scene)
            self.plot.add(self.bounding_mesh)

        # An offset is applied to the volume in build_actor, so we have to apply it here
        self.bounding_mesh.pos(np.array([-100+self.ibl_model.resolution, 0.0, 0.0]))
        #self.bounding_mesh.mapper().SetClippingPlanes(self.clipping_planes)

    def find_region(self, term):
        """
        Find a region with a substring
        :param term: Search term
        :return: List of matching entries and the corresponding mask
        """
        # ibl_model.atlas.regions is a BrainRegion object that puts a pandas dataframe
        # into separate numpy arrays (like 'name') so we work on a numpy array here
        mask = np.flatnonzero(np.char.find(self.ibl_model.atlas.regions.name.astype(str), term) != -1)
        return mask

    def get_region_names(self):
        """
        Get the region names
        :return: List
        """
        return self.ibl_model.atlas.regions.name.tolist()

    def _select(self, actor=None, controller=None, event=None,
                    camera_position=None, position=None, value=None):
        """
        Define the current object selected
        :param actor: a vtkActor
        :param controller: Controller of the given actor (optional)
        :param event: a vedo event from which we use picked3d and picked2d (we could directly use vtk)
        :param camera_position: position of the camera (optional) at selection time
        :param position: The final position computed on the volume or mesh or point or line.
            If not given, this will be automatically calculated
        .param value: The value corresponding to the point on the object. If not given, this will
            be automatically retrieved
        """
        super()._select(actor, controller, event, camera_position, position, value)
        extra_data = self.ibl_model.get_mapped_data(self.model.selection_value)
        if extra_data is None:
            return
        # This is where we retrieve neuroscience specific data about our selection
        selection = self.model.selection
        if selection == self.atlas_controller.actor or self.is_probe(selection):
            self.model.selection_related_name = extra_data.get('region_name')
            self.model.selection_related_value = extra_data.get('scalar')

    def get_selection_info(self, line_length=40, precision=5):
        """
        Get information about the current selection
        :param line_length: Region name line length after what it's word-wrapped
        :param precision: Scalar value floating precision displayed
        :return: Preformatted multiline text and a dictionary of extra data
        """
        text, data = super().get_selection_info()
        if self.atlas_controller is None:
            return text, data
        region_name = self.model.selection_related_name
        scalar = self.model.selection_related_value
        if region_name is not None:
            if isinstance(line_length, int):
                lines = textwrap.wrap(region_name, line_length, break_long_words=True)
                region_name = '\n'.join(lines)
            text += f'\nRegion: {region_name}'
        if scalar:
            text += f'\n\nScalar value: {scalar:.{precision}f}'
        return text, data

    def add_origin(self):
        """
        Add the origin on scene
        """
        self.atlas_origin = utils.Cross3DExt(self.model.origin, thickness=2, size=500)
        self.atlas_origin.lighting('off')
        self.plot.add(self.atlas_origin)
        #text_pos = self.atlas_origin.pos()+[0, 100, 0]
        #font_size = self.model.ui.font_size * 10
        #self.atlas_origin_label = vedo.Text3D('Bregma origin', pos=text_pos, s=font_size, c='k')
        #self.atlas_origin_label.followCamera()
        #self.plot.add([self.atlas_origin, self.atlas_origin_label])

    def add_many_points_test(self, positions, point_radius=2, auto_xy_rotate=True, add_to_scene=True):
        """
        Test method that validates that VTK is fast enough for displaying 10 million points interactively (and it is :)
        """
        if positions is None:
            try:
                points_path = utils.get_local_data_file_path('mouse_brain_neurons', extension='npz')
                positions = np.load(points_path, allow_pickle=True)
                positions = positions['arr_0']
            except Exception:
                # We sample a cube if you don't have the pickle file for neurons in the brain
                positions = np.random.rand(1000000, 3) * 10000
        values = np.random.rand(len(positions)) * 1.0
        point_cloud = self.add_points(positions, point_radius, values, use_origin=False, as_spheres=False)
        if auto_xy_rotate:
            point_cloud.rotateX(90)
            point_cloud.rotateZ(90)
        self.register_object(point_cloud)
        if add_to_scene:
            self.plot.add(point_cloud)
        return point_cloud

    def add_spheres(self, positions, radius=10, values=None, color_map='Accent', name='Spheres',
                    use_origin=True, add_to_scene=True, noise_amount=0, trim_outliers=True,
                    bounding_mesh=None, ibl_flip_yz=True, **kwargs):
        """
        Add new spheres
        :param positions: 3D array of coordinates
        :param radius: List same length as positions of radii. The default size is 5um, or 5 pixels
            in case as_spheres is False.
        :param values: 1D array of values, one per neuron or a time series of such 1D arrays (numpy format)
        :param color_map: A color map, it can be a color map built by IBLViewer or
            a color map name (see vedo documentation), or a list of values, etc.
        :param name: All point neurons are grouped into one object, you can give it a custom name
        :param use_origin: Whether the origin is added as offset to the given positions
        :param add_to_scene: Whether the new lines are added to scene/plot and rendered
        :param noise_amount: Amount of 3D random noise applied to each point. Defaults to 0
        :param trim_outliers: If bounding_mesh param is given, then the spheres will be trimmed,
            only the ones inside the bounding mesh will be kept
        :param bounding_mesh: A closed manifold surface mesh used for trimming segments. If None,
            the current self.bounding_mesh is used (if it exists)
        :param ibl_flip_yz: If you have an IBL data set, its 3D coordinates will be multiplied by -1
            on Y and Z axes in order to match Allen Brain Atlas volume and how it's stored by IBL.
        :return: objects.Points
        """
        axes = [1, 1, 1]
        if ibl_flip_yz:
            axes = [1, -1, -1]
            positions = np.array(positions) * [axes]
        if noise_amount is not None:
            positions += np.random.rand(len(positions), 3) * noise_amount
        link = True if add_to_scene and not trim_outliers else False
        spheres = super().add_spheres(positions, radius, values, color_map,
                                    name, use_origin, link, **kwargs)
        spheres.axes = axes
        if bounding_mesh is None:
            bounding_mesh = self.bounding_mesh
        if trim_outliers and bounding_mesh is not None:
            spheres.cutWithMesh(bounding_mesh)
            spheres.mapper().SetScalarVisibility(True)
            if add_to_scene:
                self.plot.add(spheres)
        return spheres

    def add_points(self, positions, radius=10, values=None, color_map='Accent', name='Points', screen_space=False,
                    use_origin=True, add_to_scene=True, noise_amount=0, trim_outliers=True, bounding_mesh=None,
                    ibl_flip_yz=True, **kwargs):
        """
        Add new points
        :param positions: 3D array of coordinates
        :param radius: List same length as positions of radii. The default size is 5um, or 5 pixels
            in case as_spheres is False.
        :param values: 1D array of values, one per neuron or a time series of such 1D arrays (numpy format)
        :param color_map: A color map, it can be a color map built by IBLViewer or
            a color map name (see vedo documentation), or a list of values, etc.
        :param name: All point neurons are grouped into one object, you can give it a custom name
        :param screen_space: Type of point, if True then the points are static screen-space points.
            If False, then the points are made to scale in 3D, ie you see them larger when you
            zoom closer to them, while this is not the case with screen-space points. Defaults to False.
        :param use_origin: Whether the origin is added as offset to the given positions
        :param add_to_scene: Whether the new lines are added to scene/plot and rendered
        :param noise_amount: Amount of 3D random noise applied to each point. Defaults to 0
        :param trim_outliers: If bounding_mesh param is given, then the spheres will be trimmed,
            only the ones inside the bounding mesh will be kept
        :param bounding_mesh: A closed manifold surface mesh used for trimming segments. If None,
            the current self.bounding_mesh is used (if it exists)
        :param ibl_flip_yz: If you have an IBL data set, its 3D coordinates will be multiplied by -1
            on Y and Z axes in order to match Allen Brain Atlas volume and how it's stored by IBL.
        :return: objects.Points
        """
        axes = [1, 1, 1]
        if ibl_flip_yz:
            axes = [1, -1, -1]
            positions = np.array(positions) * [axes]
        if noise_amount is not None:
            positions += np.random.rand(len(positions), 3) * noise_amount
        link = True if add_to_scene and not trim_outliers else False
        points = super().add_points(positions, radius, values, color_map, name,
                                    screen_space, use_origin, link, **kwargs)
        points.axes = axes
        if bounding_mesh is None:
            bounding_mesh = self.bounding_mesh
        if trim_outliers and bounding_mesh is not None:
            points.cutWithMesh(bounding_mesh)
            points.mapper().SetScalarVisibility(True)
            if add_to_scene:
                self.plot.add(points)
        return points

    def add_segments(self, points, end_points=None, line_width=2, values=None, color_map='Accent',
                    name='Segments', use_origin=True, add_to_scene=True, relative_end_points=False,
                    spherical_angles=None, radians=True, trim_outliers=True, bounding_mesh=None,
                    ibl_flip_yz=True):
        """
        Add a set of segments
        :param points: 3D numpy array of points of length n
        :param end_points: 3D numpy array of points of length n
        :param line_width: Line width, defaults to 2px
        :param values: 1D list of length n, for one scalar value per line
        :param color_map: A color map, it can be a color map built by IBLViewer or
            a color map name (see vedo documentation), or a list of values, etc.
        :param name: Name to give to the object
        :param use_origin: Whether the current origin (not necessarily absolute 0) is used as offset
        :param add_to_scene: Whether the new lines are added to scene/plot and rendered
        :param relative_end_points: Whether the given end point is relative to the start point. False by default,
            except is spherical coordinates are given
        :param spherical_angles: 3D numpy array of spherical angle data of length n
            In case end_points is None, this replaces end_points by finding the relative
            coordinate to each start point with the given radius/depth, theta and phi
        :param radians: Whether the given spherical angle data is in radians or in degrees
        :param trim_outliers: Whether segments are cropped by the bounding mesh
        :param bounding_mesh: A closed manifold surface mesh used for trimming segments. If None,
            the current self.bounding_mesh is used (if it exists)
        :param ibl_flip_yz: If you have an IBL data set, its 3D coordinates will be multiplied by -1
            on Y and Z axes in order to match Allen Brain Atlas volume and how it's stored by IBL.
        :return: objects.Lines
        """
        axes = [1, 1, 1]
        if ibl_flip_yz:
            axes = [1, -1, -1]
            points = np.array(points) * axes
            if end_points is not None:
                end_points = np.array(end_points) * axes
        pre_add = True if add_to_scene and not trim_outliers else False

        #lines = super().add_segments(points, end_points, line_width, values, color_map, name, use_origin,
                                    #pre_add, relative_end_points, spherical_angles, radians)
        '''
        Crazy python stuff here [WARNING]
        If the above line with super() is called, then the scope of self within super() is this one,
        which is wrong. When we are in super, self should be the parent class.
        '''
        # Copy-paste from application.Viewer.add_segments, due to above reason
        if end_points is None and spherical_angles is not None:
            relative_end_points = True
            spherical_angles = np.array(spherical_angles)
            if radians:
                end_points = spherical_angles.apply(vedo.spher2cart)
            else:
                end_points = spherical_angles.apply(utils.spherical_degree_angles_to_xyz)
            if relative_end_points:
                end_points += points
            points = np.c_[points, end_points].reshape(-1, 2, 3)
        elif end_points is not None and len(points) != len(end_points):
            n = min(len(points), len(end_points))
            logging.error(f'[add_segments() error] Mismatch between start and end points length. Only {n} segments shown.')
            points = np.c_[points[n], end_points[n]].reshape(-1, 2, 3)
        lines = super().add_lines(points, line_width, values, color_map, name, use_origin, pre_add)

        lines.axes = axes
        if bounding_mesh is None:
            bounding_mesh = self.bounding_mesh
        if trim_outliers and bounding_mesh is not None:
            lines.cutWithMesh(bounding_mesh)
            lines.mapper().SetScalarVisibility(True)
            if add_to_scene:
                self.plot.add(lines)
        return lines

    def add_lines(self, points, line_width=2, values=None, color_map='Accent', name='Lines',
                use_origin=True, add_to_scene=True, trim_outliers=True, bounding_mesh=None, ibl_flip_yz=True):
        """
        Create a set of lines with given point sets
        :param points: List of lists of 3D coordinates
        :param line_width: Line width, defaults to 2px
        :param values: 1D list of length n, for one scalar value per line
        :param color_map: A color map, it can be a color map built by IBLViewer or
            a color map name (see vedo documentation), or a list of values, etc.
        :param name: Name to give to the object
        :param use_origin: Whether the current origin (not necessarily absolute 0) is used as offset
        :param add_to_scene: Whether the new lines are added to scene/plot and rendered
        :param trim_outliers: Whether segments that are out of the brain envelope are trimmed or not. True by default
        :param bounding_mesh: A closed manifold surface mesh used for trimming lines. If None,
            the current self.bounding_mesh is used (if it exists)
        :param ibl_flip_yz: If you have an IBL data set, its 3D coordinates will be multiplied by -1
            on Y and Z axes in order to match Allen Brain Atlas volume and how it's stored by IBL.
        :return: objects.Lines
        """
        axes = [1, 1, 1]
        if ibl_flip_yz:
            axes = [1, -1, -1]
        #target =  list(points.keys()) if isinstance(points, dict) else range(len(points))
        if not isinstance(points, np.ndarray):
            '''
            This part below is to handle the numpy error that does not allow a numpy array
            to contain lists with unequal lengths.
            -> VisibleDeprecationWarning: Creating an ndarray from ragged nested sequences 
            (which is a list-or-tuple of lists-or-tuples-or ndarrays with different 
            lengths or shapes) is deprecated
            '''
            all_points = []
            indices = []
            line_id = 0
            # Possible speed improvement: use map or np.apply_along_axis
            for index in range(len(points)):
                point_set = points[index]
                point_set = np.array(point_set).astype(float)
                point_set = point_set * [axes]
                if use_origin:
                    point_set = point_set + self.model.origin
                all_points.append(point_set)
                indices.append(line_id)
                line_id += 1
            points = all_points
            if values is None:
                values = indices

        pre_add = True if add_to_scene and not trim_outliers else False
        lines = super().add_lines(points, line_width, values, color_map, name, False, pre_add)

        if bounding_mesh is None:
            bounding_mesh = self.bounding_mesh
        if trim_outliers and bounding_mesh is not None:
            lines.cutWithMesh(bounding_mesh)
            # This is the sort of thing that kills me. I was loosing all scalar colors
            # and it took me a while to see it's because of vedo's cutWithMesh that disables
            # scalar visibility!
            lines.mapper().SetScalarVisibility(True)
            if add_to_scene:
                self.plot.add(lines)
        return lines

    def add_volume(self, data=None, resolution=None, file_path=None, color_map='viridis',
                    alpha_map=None, select=False, add_to_scene=True, transpose=None):
        """
        Add a volume to the viewer with box clipping and slicing enabled by default
        :param data: Volume image data or a file_path
        :param resolution: Resoluton of the volume
        :param file_path: File path of the volume. If you don't provide an image volume data,
            then the file_path will be used to load the volume data
        :param color_map: Color map for the volume
        :param alpha_map: Alpha map for the volume. If None, it will assume that 0 values
            are transparent and maximum values are opaque
        :param select: Whether the volume is selected
        :param add_to_scene: Whether the volume is added to scene
        :param transpose: Transposition parameter. If None. nothing happens. If True,
            then the default IBL transposition is applied. You can provide your own, that is,
            a list of 3 elements to reorder the volume as desired.
        :return: VolumeController
        """
        if transpose == True:
            transpose = self.ibl_transpose
        return super().add_volume(data, resolution, file_path, color_map,
                                alpha_map, select, add_to_scene, transpose)

    def set_left_view(self):
        """
        Set left sagittal view
        """
        self.update_camera([1.0, 0.0, 0.0], self.model.Z_DOWN)

    def set_right_view(self):
        """
        Set right sagittal view
        """
        self.update_camera([-1.0, 0.0, 0.0], self.model.Z_DOWN)

    def set_anterior_view(self):
        """
        Set anterior coronal view
        """
        self.update_camera([0.0, 1.0, 0.0], self.model.Z_DOWN)

    def set_posterior_view(self):
        """
        Set posterior coronal view
        """
        self.update_camera([0.0, -1.0, 0.0], self.model.Z_DOWN)

    def set_dorsal_view(self):
        """
        Set dorsal axial view
        """
        self.update_camera([0.0, 0.0, 1.0], self.model.X_UP)

    def set_ventral_view(self):
        """
        Set ventral axial view
        """
        self.update_camera([0.0, 0.0, -1.0], self.model.X_UP)



import time
import os

import numpy as np
import pandas as pd
from scipy.ndimage import convolve

from ibllib.dsp import fcn_cosine
from ibllib.atlas import AllenAtlas

from iblviewer.mouse_brain import MouseBrainViewer
from iblviewer import utils


DIST_FCN = np.array([100, 150]) / 1e6


def compute_coverage_volume(ba=None):
    file_channels = next(utils.EXAMPLES_DATA_FOLDER.glob('channels.*.pqt'))

    print('Computing coverage volume...')
    # read the channels files and create a volume with one in each voxel containing a spike
    channels = pd.read_parquet(file_channels)
    if ba is None:
        ba = AllenAtlas(25)
    cvol = np.zeros(ba.image.shape, dtype=float)
    xyz = np.c_[channels['ml'].to_numpy(), channels['ap'].to_numpy(), channels['dv'].to_numpy()] / 1e6
    iii = ba.bc.xyz2i(xyz)
    cvol[np.unravel_index(ba._lookup(xyz), cvol.shape)] = 1

    # create the convolution kernel, 3D cosine function decreasing from 1 to 0 between bounds in DISC_FCN
    dx = ba.bc.dx
    template = np.arange(- np.max(DIST_FCN) - dx, np.max(DIST_FCN) + 2 * dx, dx) ** 2
    kernel = sum(np.meshgrid(template, template, template))
    kernel = 1 - fcn_cosine(DIST_FCN)(np.sqrt(kernel))

    # and convolve (NB: 3D FFT would probably go much faster here)
    start = time.time()
    cvol = convolve(cvol, kernel)
    print('Done in', time.time() - start, 'seconds')
    return cvol

'''
def plot_agg(ccov, axis, ax=None, ba=None, **kwargs):
    if not ax:
        ax = plt.gca()
        ax.axis('equal')
    hm = np.sum(ccov, axis=axis)
    hm[np.all(ccov == 0, axis=axis)] = np.nan
    # hm[np.all(ccov == 0, axis=axis)] = np.nan
    if axis == 0:
        axextent = 1  # coronal
        hm = hm.T
    elif axis == 1:
        axextent = 0  # sagittal
        hm = hm.T
    elif axis == 2:
        axextent = 2  # horizontal
    ax.imshow(hm, extent=ba.extent(axis=axextent), **kwargs)
    return hm
'''


def main():
    viewer = MouseBrainViewer()
    resolution = 50
    #viewer.initialize(resolution=resolution, mapping='Allen-lr', embed_ui=True, dark_mode=False, add_atlas=False, add_dwi=True, dwi_color_map='viridis')
    viewer.initialize(resolution=resolution, mapping='Allen-lr', embed_ui=True, dark_mode=False)

    ba = viewer.ibl_model.atlas
    file_path = utils.EXAMPLES_DATA_FOLDER.joinpath(f'./ncov_{resolution}.npz')
    if os.path.exists(str(file_path)):
        ncov = np.load(str(file_path))['arr_0']
    else:
        cvol = compute_coverage_volume(ba)
        ncov = cvol.copy()
        ncov[ncov < 0] = -1
        ncov += 1
        print('Done computing volume with range', np.min(ncov), np.max(ncov))
        np.savez_compressed(str(file_path), ncov)

    cov_vol = viewer.add_volume(ncov, resolution, color_map='viridis', transpose=True, select=True) #, alpha_map=[0, 0.5, 1]
    cov_vol.set_opacity(1.0)
    viewer.show().close()
    '''
    ncov[ncov == -1] = np.nan

    plt.figure(), plot_agg(ncov, 2, ax=ba.plot_hslice(-.002), ba=ba, alpha=0.5)
    plt.figure(), plot_agg(ncov, 1, ax=ba.plot_sslice(0), ba=ba, alpha=0.5)
    plt.figure(), plot_agg(ncov, 0, ax=ba.plot_cslice(0), ba=ba, alpha=0.5)
    '''

if __name__ == '__main__':
    main()


import numpy as np
import pandas as pd
import pickle
import argparse

from iblviewer.launcher import IBLViewer
from iblviewer import utils

class DataViewer():

    def __init__(self, file_path=None, aggregator='median', grouper='acronym', color_map='viridis'):
        """
        Constructor
        """
        self.file_path = file_path
        self.aggregator = aggregator
        self.color_map = color_map
        self.grouper = grouper
        self.df = None
        self.aggregated_df = None
        self.min_value = None
        self.max_value = None
        self.load_data(file_path)

    def load_data(self, file_path=None, silent=True):
        if file_path is None:
            test_data = './stimonR_top10_rawpoints.p'
            file_path = str(utils.EXAMPLES_DATA_FOLDER.joinpath(test_data))
            self.file_path = file_path
        
        df = pickle.load(open(file_path, 'rb'))
        
        # For testing data given by Guido
        # df['r_over_chance'] = df['r_prior'] - df['r_prior_null']
        # filtered_df = df.groupby('region').median()['r_over_chance']

        raw_df = df['rawpoints']
        copy_df = raw_df.copy()
        agg_df = copy_df.groupby(self.grouper).agg({'value': self.aggregator})
        agg_df.dropna(inplace=True)
        self.min_value = float(np.amin(agg_df, axis=0).to_numpy()[0])
        self.max_value = float(np.amax(agg_df, axis=0).to_numpy()[0])

        if not silent:
            print('Min prior value ' + str(self.min_value))
            print('Max prior value ' + str(self.max_value))

        self.df = raw_df
        self.df.dropna(inplace=True)
        self.df.sort_values(by='acronym', key=lambda col: col.str.lower())
        self.aggregated_df = agg_df
        
    #pLeft_iti_scores_n_gt_50
    # Data given by Berk Gerçek, International Brain Laboratory
    def get_scalars_map(self, viewer):
        """
        Process priors data and get color map and scalar values
        """
        scalars_map = [None]*viewer.ibl_model.get_num_regions()

        # This code is to be modified if you have split data for left and right hemispheres
        # The concept is pretty simple: scalars_map is a 1D list that maps to brain regions.
        # With the lateralized brain mapping, the standard region id in Allen CCF is negated
        # on the right hemisphere. 
        # Currently this code uses standard acronym lookup, which yields a region on both
        # hemispheres. The value you assign to an acronym will thus be mirrored.

        # Or for i in range(0, len(df)): which preserves data types
        for acronym, row in self.aggregated_df.iterrows():
            value = row.iloc[0]
            if value is None:
                continue
            region_ids, row_ids = viewer.ibl_model.get_region_and_row_id(acronym)
            if region_ids is None:
                print('Acronym', acronym, 'was not found in Atlas')
                continue
            for r_id in range(len(region_ids)):
                region_id = region_ids[r_id]
                row_id = row_ids[r_id]
                if region_id is None:
                    print('Error, could not find acronym', acronym, '...ignoring it)')
                    continue
                if row_id == 0: #or value.isnull().values.any():
                    # We ignore void acronym (which is equal to row_id 0) on Allen Mouse CCF v3
                    continue
                #print('Setting row', int(row_id), 'with value', value)
                scalars_map[int(row_id)] = value
        return scalars_map
        
    def on_viewer_initialized(self, viewer):
        """
        Method called when the viewer is initialized and ready to accept further code by the user.
        In this case we map prior data to the Allen Mouse brain atlas.
        :param viewer: MouseBrainViewer instance (mandatory)
        """
        scalars_map = self.get_scalars_map(viewer)
        # Ensure we have the right selection, here 0 is the first registered object: the atlas
        viewer.select(-1)
        viewer.assign_scalars(scalars_map, [self.min_value, self.max_value], self.color_map)

    def on_statistics_update(self, statistics, viewer):
        """
        Method called when statistics are updated. Here we use a scatter plot.
        :param statistics: MplCanvas instance (mandatory)
        :param viewer: MouseBrainViewer instance (mandatory)
        """
        # Clear previous plot (nothing visible appears until you call draw())
        statistics.axes.clear()

        # Prepare new one
        agg_df = self.aggregated_df
        statistics.axes.set_xlabel('Brain regions (none selected)')
        # First scatter all values
        statistics.axes.scatter(self.df.acronym, self.df.value, alpha=0.2, s=8)
        statistics.axes.set_xticks([''])
        # There are multiple ways to retrieve the acronym, here's one
        acronyms = agg_df.index[agg_df.value == viewer.model.selection_related_value].tolist()
        if acronyms is None or len(acronyms) < 1:
            # Do not forget to draw before we return!
            statistics.draw()
            return

        # Then scatter the values on top of the previous scatter with the values from the selected region
        for acronym in acronyms:
            #selected_index = self.df.index[self.df.acronym == acronym]
            selected_data = self.df.value[self.df.acronym == acronym]
            #selected_index.dropna(inplace=True)
            selected_data.dropna(inplace=True)
            statistics.axes.scatter([acronym]*selected_data.size, selected_data, color='yellow', s=32)
        
        statistics.axes.set_xlabel(f'Selected region: {acronyms}')
        # At last, draw the result
        statistics.draw()


def main():
    # More parsing options are added in parse_args() method below.
    # -> Please check that you don't override any existing argument name!
    parser = argparse.ArgumentParser(description='International Brain Viewer based on VTK')
    parser.add_argument('-f', dest='file_path', type=str, default=None, 
    help='File path to your data. None will load a default example.')
    parser.add_argument('-g', dest='grouper', type=str, default='acronym', 
    help='Grouper type. Defaults to "acronym" column.')
    parser.add_argument('-a', dest='aggregator', type=str, default='median', 
    help='Aggregator type. Defaults to median.')

    iblviewer = IBLViewer()
    # First retrieve command-line arguments (default ones + custom ones above)
    args = iblviewer.parse_args(parser)

    # Then update our custom class with those arguments
    data = DataViewer(args.file_path, args.aggregator, args.grouper, args.color_map)

    # Finally, launch the UI and 3D viewer
    iblviewer.launch(data.on_viewer_initialized, data.on_statistics_update, args)


if __name__ == '__main__':
    main()


# This example connects to IBL database and fetches
# all insertion probes
try:
    from oneibl.one import ONE
except ImportError:
    from one.api import ONE
import numpy as np

import argparse
from iblviewer.launcher import IBLViewer


class ProbeData:

    def __init__(self, as_segments=False, line_width=2, trim_outliers=True):
        """
        Constructor
        :param viewer: The IBLViewer controller
        :param one_connection: The "one" connection to IBL server
        :param as_segments: Whether insertion probes should be reduced to straight lines (segments)
        :param trim_outliers: Whether you want the lines to be cut when they're out of the brain
        :param with_labels: Whether labels should be added to the lines
        """
        print('You need valid credentials and access to IBL server in order to run this example.') 
        self.one_connection = ONE(base_url="https://alyx.internationalbrainlab.org")
        self.as_segments = as_segments
        self.line_width = line_width
        self.trim_outliers = trim_outliers

    def get_bwm_ins_alyx(self, one):
        """
        Return insertions that match criteria :
        - project code
        - session QC not critical (TODO may need to add probe insertion QC)
        - at least 1 alignment
        - behavior pass
        :param one: "one" connection handler
        :return:
        ins: dict containing the full details on insertion as per the alyx rest query
        ins_id: list of insertions eids
        sess_id: list of (unique) sessions eids
        """
        ins = one.alyx.rest('insertions', 'list',
                            provenance='Ephys aligned histology track',
                            django='session__project__name__icontains,ibl_neuropixel_brainwide_01,'
                                'session__qc__lt,50,'
                                'json__extended_qc__alignment_count__gt,0,'
                                'session__extended_qc__behavior,1')
        
        ins_ids = [item['id'] for item in ins]
        sess_id = [item['session_info']['id'] for item in ins]
        # Here's what's in 'json':
        # dict_keys(['qc', 'n_units', 'xyz_picks', 'extended_qc', 'drift_rms_um', 'firing_rate_max', 'n_units_qc_pass', 
        # 'amplitude_max_uV', 'firing_rate_median', 'amplitude_median_uV', 'whitening_matrix_conditioning'])
        positions = []
        for item in ins:
            picks = item['json'].get('xyz_picks', [])
            positions.append(picks)
        sess_id = np.unique(sess_id)
        return positions, ins_ids


    def get_picks_mean_vectors(self, xyz_picks, extent=3):
        """
        Get a mean vector from picks coordinates
        :param xyz_picks: Dictionary xyz picks, the key being the identifier for that data set
        :param extent: Number of points to take from start and end for mean computation of end points
        :return: List of varying lists of 3D points and a list of line ids
        """
        vectors = []
        ids = []
        # Mean between first and last three picks
        for ins_id in range(len(xyz_picks)):
            raw_picks = xyz_picks[ins_id]
            end_pt = np.mean(raw_picks[-extent:], axis=0)
            start_pt = np.mean(raw_picks[:extent], axis=0)
            vectors.append([start_pt, end_pt])
            ids.append(ins_id)
        return vectors, ids


    def on_viewer_initialized(self, viewer):
        """
        Add insertion probe vectors
        :param viewer: Viewer instance
        """
        lines_data, line_ids = self.get_bwm_ins_alyx(self.one_connection)
        if self.as_segments:
            segments_data, segment_ids = self.get_picks_mean_vectors(lines_data)
            line_ids = np.array(line_ids)
            segment_ids = line_ids[segment_ids]
            lines = viewer.add_segments(segments_data, line_width=self.line_width, 
                                        add_to_scene=True, trim_outliers=self.trim_outliers)
        else:
            lines = viewer.add_lines(lines_data, line_width=self.line_width,
                                    add_to_scene=True, trim_outliers=self.trim_outliers)
        return lines


def str2bool(v):
    if isinstance(v, bool):
        return v
    elif v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def main():
    # More parsing options are added in parse_args() method below.
    # -> Please check that you don't override any existing argument name!
    parser = argparse.ArgumentParser(description='International Brain Viewer based on VTK')
    parser.add_argument('-seg', dest='segments', type=str2bool, default=False, 
    help='Whether line probes are simplified to segments. Defaults to 0 (False)')

    iblviewer = IBLViewer()
    # First retrieve command-line arguments (default ones + custom ones above)
    args = iblviewer.parse_args(parser)

    pb = ProbeData(args.segments)
    # Now start the viewer and add points when it's initialized
    iblviewer.launch(pb.on_viewer_initialized, None, args)
    '''
    # If you'd like, you can also directly import the relevant viewer and use it as follows:
    from iblviewer.mouse_brain import MouseBrainViewer
    viewer = MouseBrainViewer()
    viewer.initialize(resolution=25, embed_ui=True)
    add_insertion_probes(viewer, one_connection, as_segments=False, line_width=5)
    viewer.show()
    '''


if __name__ == '__main__':
    main()



from collections import OrderedDict


class Collection(OrderedDict):

    current = None
    current_key_id = 0
    targets = []

    def get_current_key(self):
        """
        Get the current key that maps to the current object
        """
        return self.get_keys()[self.current_key_id]

    def set_current(self, target):
        """
        Set the current data
        :param key_or_id: A valid key or an int or the data
        object itself
        """
        keys = self.get_keys()
        if target in self:
            key = target
        elif isinstance(target, int):
            key_id = target
            try:
                key = keys[key_id]
            except Exception:
                return
        else:
            for key in self:
                if self[key] == target:
                    break
        
        key_id = 0
        for other_key in keys:
            if other_key == key:
                break
            key_id += 1
        
        # This is for when you need to know where is the key in the array
        self.current_key_id = key_id
        self.current = self.get(key)

    def next(self, loop=False):
        """
        Set the next data as current one
        :param loop: Whether looping is enabled,
            which means that if next value is out of range,
            we start back at 0
        """
        keys = self.get_keys()
        new_id = self.current_id + 1
        if new_id > len(keys) - 1:
            new_id = 0 if loop else len(keys) - 1
        self.set_current(new_id)
        return new_id

    def previous(self, loop=False):
        """
        Set the previous data as current one
        :param loop: Whether looping is enabled,
            which means that if previous value is out of range,
            we go to -1 in a backward loop
        """
        keys = self.get_keys()
        new_id = self.current_id - 1
        if new_id < 0:
            new_id = len(keys) - 1 if loop else 0
        self.set_current(new_id)
        return new_id

    def store(self, data, data_id=None, replace_existing=True, set_current=False):
        """
        Store a data
        :param data: data instance
        :param data_id: data id, a unique string if possible.
            If it's not a unique string, it will be appended a
            number so that it's unique.
        :param replace_existing: Whether any existing data with the same id is replaced or not
        :param set_current: Whether the newly stored data is set as the current one
        :return: The new data_id or None
        """
        if data_id is None or (data_id in self and not replace_existing):
            data_id = self.get_new_name(data)
        if data_id in self and not replace_existing:
            return
        self[data_id] = data
        if set_current or len(self) == 1:
            self.set_current(data_id)
        return data_id

    def get_or_create(self, data_id, data_class=None):
        """
        Get a data from a dictionary of data or create it if none found.
        This method works only with datas that have a name (str) property.
        :param data_id: Either a data name or its id
        :param data_class: The class of the data
        :return: A data of type data_class
        """
        name = data_id
        if isinstance(data_id, int):
            name = self.get_keys()[data_id]
        data = self.get(name)
        if data is None and data_class is not None:
            if isinstance(data_id, str):
                data = data_class(name=data_id)
            else:
                data = data_class()
                data_id = self.get_new_name(data)
            self[data_id] = data
        return data

    def get_keys(self):
        """
        Get all data ids
        """
        return list(self.keys())

    def find_keys(self, id_or_subid):
        """
        Get all ids/keys that have the given param as a substring
        """
        keys = self.get_keys()
        found_ones = []
        for key in keys:
            if id_or_subid in key: #key could be an array or any iterable too
                found_ones.append(key)
        return found_ones

    def get_new_name(self, obj):
        """
        Get a new name derived from an object's class name
        :return: String
        """
        return f'{obj.__class__.__name__}{len(self)}'

    def get_name(self, *args):
        """
        Get given arguments separated by underscores as a single string
        :return: String
        """
        return '_'.join(args)



import argparse
from cloudvolume import CloudVolume
from iblviewer.launcher import IBLViewer


class HumanBrainData():

    def __init__(self):
        self.volume = None
        self.cloud_volume = None
        self.atlas = 'classif'
        self.lod = 3
        self.resolution = None
        self.color_map = None

    def load_volume(self, atlas='classif', lod=3):
        url = 'precomputed://https://neuroglancer.humanbrainproject.eu/'
        url += 'precomputed/BigBrainRelease.2015/' + atlas
        print('About to load precomputed volume', url)
        volume = CloudVolume(url, mip=lod)
        print('Metadata for lod', lod, ':', volume.scale)
        image = volume[:, :, :]
        self.volume = image.flatten().reshape(volume.scale['size'])
        self.cloud_volume = volume
        self.resolution = volume.scale['resolution'][0]

    def on_viewer_initialized(self, viewer):
        viewer.add_volume(self.volume, self.resolution, None, self.color_map, select=True)
        viewer.show()


def main():
    parser = argparse.ArgumentParser(description='Human Brain Atlas')
    parser.add_argument('-l', dest='lod', type=int, default=3, 
    help='LOD value')
    parser.add_argument('-v', dest='volume', type=str, default='classif', 
    help='Volume type, either 8bit or classif')
    parser.add_argument('-c', dest='color_map', type=str, default='viridis', 
    help='Color map')

    iblviewer = IBLViewer()
    args = iblviewer.parse_args(parser)

    hb = HumanBrainData()
    hb.color_map = args.color_map
    hb.load_volume(args.volume, args.lod)

    iblviewer.launch(hb.on_viewer_initialized)


if __name__ == '__main__':
    main()


from datetime import datetime
import numpy as np
import os
import glob
from pathlib import Path

from vtk.util.numpy_support import numpy_to_vtk, vtk_to_numpy, numpy_to_vtkIdTypeArray
import vtk
import vedo
import math
#import trimesh


ROOT_FOLDER = Path(__file__).parent.parent
ASSETS_FOLDER = ROOT_FOLDER.joinpath('./iblviewer_assets')
FONTS_FOLDER = ASSETS_FOLDER.joinpath('./fonts')
EXAMPLES_FOLDER = ROOT_FOLDER.joinpath('./iblviewer_examples')
EXAMPLES_DATA_FOLDER = ROOT_FOLDER.joinpath('./iblviewer_examples/data')


def get_type(element):
    """
    Get the type of object as a string
    :return: String
    """
    return str(element.__class__.__name__).lower()


def get_unique_name(collection, name, spacer='_'):
    """
    Create a unique key for a collection by appending numbers when entries exist
    :param collection: A list, collection, array, ...
    :param name: Name (for instance 'Points')
    :param spacer: Spacer char
    :return: New name, for instance 'Points_4'
    """
    similar_ones = []
    max_value = 0
    if name not in collection:
        return name
    for key in collection:
        if name in key:
            similar_ones.append(key)
            if spacer in key:
                value = key.split(spacer)[1]
                max_value = max(int(value), max_value)
    value = max(len(similar_ones), max_value)
    return f'{name}{spacer}{value}'
    

def numpy2vtk(arr, dtype=None, deep=True, name=""):
    """
    Convert a numpy array into a vtkDataArray
    :param arr: Array
    :param dtype: Data type. Allows to set a specific data type to the VTK array
    :param deep: Whether a deep copy is made. Defaults to True
    :param name: Name of the array
    """
    if arr is None:
        return None
    arr = np.ascontiguousarray(arr)
    if dtype is not None and dtype!='id':
        arr = arr.astype(dtype)

    if dtype and dtype=='id':
        varr = numpy_to_vtkIdTypeArray(arr.astype(np.int64), deep=deep)
    else:
        varr = numpy_to_vtk(arr, deep=deep)
    if name:
        varr.SetName(name)
    return varr


def spherical_degree_angles_to_xyz(radius, theta, phi):
    """
    Convert spherical degree angles to XYZ coordinates
    :param radius: Radius
    :param theta: Theta angle value in degrees
    :param phi: Phi angle value in degrees
    :return: List of 3 coordinates
    """
    return vedo.spher2cart(radius, theta / 180 * math.pi, phi / 180 * math.pi) 


def pick_object(plot, event_name=None, priority=None, cid=None):
    """
    Pick an object
    """
    x, y = plot.interactor.GetEventPosition()
    plot.renderer = plot.interactor.FindPokedRenderer(x, y)
    if not plot.picker:
        plot.picker = vtk.vtkPropPicker()
    plot.picker.PickProp(x, y, plot.renderer)
    plot.picked2d = (x,y)
    xp, yp = plot.interactor.GetLastEventPosition()
    actor = plot.picker.GetProp3D()
    delta3d = np.array([0,0,0])
    picked3d = None
    picker = plot.picker
    if actor is None:
        # Ok, this is tricky. I found out that vtkPropPicker, even
        # if it optimized, can fail at detecting a simple mesh
        # so we use the vtkPicker as fall back plan
        picker = vtk.vtkPicker()
        picker.Pick(x, y, 0.0, plot.renderer)
        actor = picker.GetProp3D()
    if actor is not None:
        picked3d = np.array(picker.GetPickPosition())
        if isinstance(actor, vedo.Mesh):
            # There is a bug with transparent objects or objects that do not have ForceOpaqueOn()
            # which prevents picked3d from being valid so we have to use another picking method
            cell_picker = vtk.vtkCellPicker()
            cell_picker.Pick(x, y, 0.0, plot.renderer)
            if cell_picker.GetProp3D() == actor:
                picked3d = np.array(cell_picker.GetPickPosition())
        try:
            if actor.picked3d is not None:
                delta3d = picked3d - actor.picked3d
            actor.picked3d = picked3d
        except AttributeError:
            return
    else:
        actor = plot.picker.GetActor2D()

    dx, dy = x-xp, y-yp

    event_dict = vedo.utils.dotdict({
        "name": event_name,
        "id": cid,
        "priority": priority,
        "at": plot.renderers.index(plot.renderer),
        "actor": actor,
        "picked3d": picked3d,
        "keyPressed": plot.interactor.GetKeySym(),
        "picked2d": (x,y),
        "delta2d": (dx, dy),
        "angle2d": np.arctan2(dy,dx),
        "speed2d": np.sqrt(dx*dx+dy*dy),
        "delta3d": delta3d,
        "speed3d": np.sqrt(np.dot(delta3d,delta3d)),
        "isPoints":   isinstance(actor, vedo.Points),
        "isMesh":     isinstance(actor, vedo.Mesh),
        "isAssembly": isinstance(actor, vedo.Assembly),
        "isVolume":   isinstance(actor, vedo.Volume),
        "isPicture":  isinstance(actor, vedo.Picture),
        "isActor2D":  isinstance(actor, vtk.vtkActor2D)
    })
    return event_dict


def add_callback(plot, event_name, func, priority=0.0):
    """
    Modified function from vedo. The issue is that the way vedo (and pyvista for that matter)
    is structured is that it helps using vtk but sometimes hinders using it with code that makes
    assumptions we don't want.

    Add a function to be executed while show() is active.
    Information about the event can be acquired with method ``getEvent()``.
    Return a unique id for the callback.
    The callback function (see example below) exposes a dictionary
    Frequently used events are:
        - KeyPress, KeyRelease: listen to keyboard events
        - LeftButtonPress, LeftButtonRelease: listen to mouse clicks
        - MiddleButtonPress, MiddleButtonRelease
        - RightButtonPress, RightButtonRelease
        - MouseMove: listen to mouse pointer changing position
        - MouseWheelForward, MouseWheelBackward
        - Enter, Leave: listen to mouse entering or leaving the window
        - Pick, StartPick, EndPick: listen to object picking
        - ResetCamera, ResetCameraClippingRange
        - Error, Warning
        - Char
        - Timer
    Check the complete list of events here:
        https://vtk.org/doc/nightly/html/classvtkCommand.html
    """
    if not plot.interactor:
        return None

    # Processing names is removed from vedo function
    # Also the whole thing is refactored with improved picking
    def wrapper(iren=None, event_name=None):
        event_dict = pick_object(plot, event_name, priority, cid)
        func(event_dict)

    cid = plot.interactor.AddObserver(event_name, wrapper, priority)
    return cid


def get_file_name(file_name, extension):
    """
    Get full file name
    :param file_name: File name without extension
    :param extension: File extension
    :return: File name with extension
    """
    if str(file_name).endswith(extension):
        full_file_name = str(file_name)
    else:
        full_file_name = str(file_name) + '.' + str(extension)
    return full_file_name


def get_local_data_file_path(file_name, extension, sub_folder=''):
    """
    Get data path
    :param file_name: File name without extension
    :param extension: File extension
    :return: File path
    """
    return ASSETS_FOLDER.joinpath(sub_folder, get_file_name(file_name, extension))


def get_surface_mesh_path(file_name, meshes_path=None, extension='ply', default_meshes_path=None):
    """
    Get a surface mesh file path
    :param file_name: File name without extension
    :param meshes_path: Folder path. If None given, this method will look into the data folder of iblviewer
    :param extension: File extension
    :param default_meshes_path: Fallback local or remote path
    :return: Full mesh file path
    """
    if meshes_path is None:
        region_mesh_path = str(get_local_data_file_path(file_name, extension, 'surfaces'))
        if not os.path.exists(region_mesh_path):
            if default_meshes_path is not None:
                region_mesh_path = default_meshes_path
            else:
                region_mesh_path = 'https://raw.github.com/int-brain-lab/iblviewer/main/iblviewer_assets/surfaces/'
            region_mesh_path += get_file_name(file_name, extension)
    else:
        region_mesh_path = str(os.path.join(meshes_path, get_file_name(file_name, extension)))
    return region_mesh_path


def load_surface_mesh(file_name, meshes_path=None, extension='ply'):
    """
    Load a surface mesh with vedo.
    :param file_name: File name without extension
    :param meshes_path: Folder path. If None given, this method will look into the data folder of iblviewer
    :param extension: File extension
    :return: Mesh or None if path is invalid
    """
    file_path = get_surface_mesh_path(file_name, meshes_path, extension)
    if file_path.startswith('https') or os.path.exists(file_path):
        return vedo.load(file_path)


def change_file_name(file_path, prefix=None, name=None, suffix=None):
    """
    Change the file name from the given file path
    :param file_path: Input file path
    :param prefix: Prefix to the file name
    :param name: Whether a new name is set instead of the current name.
        If None, the current file name is used.
    :param suffix: Suffix to the file name
    :return: New file path
    """
    path, file_name, extension = split_path(file_path)
    if prefix is None:
        prefix = ''
    if suffix is None:
        suffix = ''
    if name is None or name == '' or not isinstance(name, str):
        name = file_name
    return os.path.join(path, prefix + name + suffix + extension)


def split_path(path):
    """
    Split any given file path to folder path, file name and extension
    :return: Folder path, file name and extension
    """
    base_name = os.path.basename(path)
    file_name, extension = os.path.splitext(base_name)
    return path[:-len(base_name)], file_name, extension


def time_diff(t):
    """
    Get a time difference in seconds
    :param t: Time
    :return: Number of seconds
    """
    now  = datetime.now()
    duration = now - t
    return duration.total_seconds()


def recompute_normals(target):
    pdnorm = vtk.vtkPolyDataNormals()
    pdnorm.SetInputData(target)
    pdnorm.ComputePointNormalsOn()
    pdnorm.ComputeCellNormalsOn()
    pdnorm.FlipNormalsOff()
    pdnorm.ConsistencyOn()
    pdnorm.Update()
    return pdnorm.GetOutput() #self._data


def get_actor_center(actor):
    """
    Get the absolute center position of an actor
    :param actor: VTK actor
    :return: 3d array
    """
    try:
        if isinstance(actor, vedo.Volume):
            return actor.center() + actor.pos()
        else:
            return actor.centerOfMass() + actor.pos() # TODO: check that this is necessary (adding pos)
    except Exception as e:
        raise e


def get_actor_bounds(actor):
    """
    Get the bounds of an actor as xmin, xmax, ymin, ymax, zmin, zmax
    :param actor: VTK actor
    :return: Array with 6 values
    """
    if actor is None:
        return
    try:
        if isinstance(actor, vedo.Volume):
            d = actor.dimensions() * actor.spacing()
            c = get_actor_center(actor)
            return c[0] - d[0], c[0] + d[0], c[1] - d[1], c[1] + d[1], c[2] - d[2], c[2] + d[2]
        else:
            return actor.bounds()
    except Exception as e:
        raise e


def get_actor_dimensions(actor):
    """
    Get the dimensions of an actor
    :param actor: VTK actor
    :return: 3d array
    """
    if actor is None:
        return
    try:
        if isinstance(actor, vedo.Volume):
            return actor.dimensions() * actor.spacing()# equivalent to self.model.resolution
        else:
            xmin, xmax, ymin, ymax, zmin, zmax = actor.bounds()
            return np.array([xmax - xmin, ymax - ymin, zmax - zmin])
    except Exception as e:
        raise e


def get_bounding_planes(actor):
    """
    Get bounding planes for an actor
    :param actor: VTK actor
    :return: vtkPlanes
    """
    planes = vtk.vtkPlanes()
    planes.SetBounds(actor.GetBounds())
    return planes


def get_planes_bounds(planes):
    """
    Get the bounding box coordinates of a series of planes.
    [WARNING] Only works for six planes (box mode) at the moment
    :param planes: vtkPlaneCollection
    :return: 6 values
    """
    origins = list()
    for p_id in range(planes.GetNumberOfItems()):
        plane = planes.GetItem(p_id)
        origin = np.array(plane.GetOrigin())
        origins.append(origin)
    # We don't want zeros to be accounted for so we select planes of interest
    # First x planes, then y planes, then z ones. To be improved/generalized.
    origins = np.array(origins)
    mi_x = np.min(origins[:2], axis=0).tolist()
    ma_x = np.max(origins[:2], axis=0).tolist()
    mi_y = np.min(origins[2:4], axis=0).tolist()
    ma_y = np.max(origins[2:4], axis=0).tolist()
    mi_z = np.min(origins[4:6], axis=0).tolist()
    ma_z = np.max(origins[4:6], axis=0).tolist()
    return mi_x[0], ma_x[0], mi_y[1], ma_y[1], mi_z[2], ma_z[2]


def get_transformation_matrix(origin, normal):
    """
    Get transformation matrix for a plane given by its origin and normal
    :param origin: Origin 3D vector
    :param normal: Normal 3D vector
    :return: Matrix and Translation
    """
    newaxis = vedo.utils.versor(normal)
    initaxis = (0, 0, 1)
    crossvec = np.cross(initaxis, newaxis)
    angle = np.arccos(np.dot(initaxis, newaxis))
    T = vtk.vtkTransform()
    T.PostMultiply()
    T.RotateWXYZ(np.rad2deg(angle), crossvec)
    T.Translate(np.array(origin))
    M = T.GetMatrix()
    return M, T


def set_clipping_planes(target, planes, flip_normals=False):
    """
    Clip the volume and move the slicing planes according the given planes
    :param target: vedo.Mesh (can be iblviewer.objects.Points or similar)
    :param planes: vtkPlanes
    :param flip_normals: Whether clipping normals are flipped
    """
    if not isinstance(planes, vtk.vtkPlanes):
        bounds = planes
        planes = vtk.vtkPlanes()
        planes.SetBounds(bounds)
    if flip_normals:
        for plane_id in range(planes.GetNumberOfPlanes()):
            plane = planes.GetPlane(plane_id)
            plane.SetNormal(np.array(plane.GetNormal())*-1)
    target.GetMapper().SetClippingPlanes(planes)


def probe(plot, target, widget=None, interaction_callback=None, point1=None, point2=None, 
            place_factor=1, handle_size=0.005, color=None):
    """
    Initializes a line widget on the given target
    :param plot: vtk plot
    :param target: Target object
    :param widget: Existing widget. In case a valid vtkLineWidget is given,
    it will be used and modified directly
    :param interaction_callback: Function that will be called every
    time there is an interaction with the widget. That's where
    you set the clipping planes to the object for instance
    :param point1: Initial position of point 1
    :param point2: Initial position of point 2
    :param place_factor: see vtkBoxWidget.setPlaceFactor()
    :param handle_size: set the relative handle size, see vtkBoxWidget.SetHandleSize()
    :param color: Color of the line
    :return: vtkLineWidget
    """
    existing = isinstance(widget, vtk.vtkLineWidget)
    if not existing:
        widget = vtk.vtkLineWidget()
        widget.SetInteractor(plot.interactor)
        widget.SetPlaceFactor(place_factor)
        widget.SetHandleSize(handle_size)
    widget.SetInputData(target.GetMapper().GetInput())
    
    if color is None:
        color = [0.5, 0.5, 0.5]
    widget.GetSelectedLineProperty().SetColor(*color)
    #widget.GetSelectedLineProperty().SetOpacity(0.7)
    if point1 is not None and point2 is not None:
        widget.PlaceWidget()
        widget.SetPoint1(*point1)
        widget.SetPoint2(*point2)
    else:
        widget.PlaceWidget(target.GetBounds())

    if interaction_callback is not None:
        widget.RemoveObservers('InteractionEvent')
        widget.AddObserver('InteractionEvent', interaction_callback)

    plot.interactor.Render()
    widget.On()
    plot.widgets.append(widget)
    return widget


def box_widget(plot, target, interaction_callback=None, place_factor=1, 
                handle_size=0.005, outline_color=None):
    """
    Initializes a box widget on the given target
    :param plot: vtk plot
    :param target: Target object
    :param interaction_callback: Function that will be called every
        time there is an interaction with the widget. That's where
        you set the clipping planes to the object for instance
    :param place_factor: see vtkBoxWidget.setPlaceFactor()
    :param handle_size: set the relative handle size, see vtkBoxWidget.SetHandleSize()
    :param outline_color: Color of the outline
    :return: vtkBoxWidget
    """
    widget = vtk.vtkBoxWidget()
    widget.SetInteractor(plot.interactor)
    widget.SetPlaceFactor(place_factor)
    widget.SetHandleSize(handle_size)
    # TODO: handle the update event in volumes in order to choose
    # the best method, either axis-aligned slicing when normals
    # are axis-aligned, or slicing with vtkImageReslice. 
    # Both functionalities are available in VolumeView already.
    widget.RotationEnabledOff()
    widget.ScalingEnabledOn()
    widget.TranslationEnabledOn()
    widget.SetInputData(target.GetMapper().GetInput())
    plot.cutterWidget = widget
    
    # Only valid for vtkBoxWidget
    widget.OutlineCursorWiresOn()
    widget.InsideOutOn()
    widget.GetSelectedOutlineProperty().SetColor(1, 0, 1)
    if outline_color is None:
        outline_color = [0.5, 0.5, 0.5]
    widget.GetOutlineProperty().SetColor(*outline_color)
    #widget.GetOutlineProperty().SetOpacity(0.7)

    #widget.SetRepresentationToOutline()
    existing_planes = target.GetMapper().GetClippingPlanes()
    if existing_planes is not None:
        try:
            bounds = get_planes_bounds(existing_planes)
            widget.PlaceWidget(bounds)
        except Exception:
            msg = '[Warning] Object ' + target.name + ' does not have six clipping planes.'
            msg += 'Placing widget is not supported atm in this configuration.'
            print(msg)
            widget.PlaceWidget(target.GetBounds())
    else:
        widget.PlaceWidget(target.GetBounds())
    
    def clip_target(widget=None, event=None):
        """
        Clip the target with the current box widget
        """
        if widget is None:
            return
        clipping_planes = vtk.vtkPlanes()
        widget.GetPlanes(clipping_planes)
        target.GetMapper().SetClippingPlanes(clipping_planes)

    if interaction_callback is None:
        interaction_callback = clip_target
    
    widget.AddObserver('InteractionEvent', interaction_callback)
    plot.interactor.Render()
    widget.On()
    plot.widgets.append(widget)
    return widget


def update_scalar_bar(sb, lut, use_alpha=False, nan_color=None):
    """
    Update a scalar bar with a new LUT
    :param sb: vtkScalarBarActor
    :param lut: vtkLookupTable
    :param use_alpha: whether alpha is used in the scalar bar
    """
    if sb.GetLookupTable() == lut:
        return
    sb.SetLookupTable(lut)
    sb.SetUseOpacity(use_alpha)
    sb.SetDrawFrame(0)
    sb.SetDrawBackground(0)
    if lut.GetUseBelowRangeColor():
        sb.DrawBelowRangeSwatchOn()
        sb.SetBelowRangeAnnotation('')
    if lut.GetUseAboveRangeColor():
        sb.DrawAboveRangeSwatchOn()
        sb.SetAboveRangeAnnotation('')
    if nan_color is not None:#lut.GetNanColor() != (0.5, 0.0, 0.0, 1.0):
        lut.SetNanColor(*nan_color)
        sb.DrawNanAnnotationOn()
        sb.SetNanAnnotation('nan')


def add_scalar_bar(lut, pos=(0.8, 0.05), font_color=[0, 0, 0], title="", titleYOffset=15, titleFontSize=12,
                    size=(None,None), nlabels=None, horizontal=False, use_alpha=False):
    """
    Create a new 2D scalar bar. This is a modified method from vedo.addons.addScalarBar
    :param lut: Color map LUT
    :param list pos: fractional x and y position in the 2D window
    :param list size: size of the scalarbar in pixel units (width, heigth)
    :param int nlabels: number of numeric labels to be shown
    :param bool use_alpha: retain trasparency in scalarbar
    :param bool horizontal: show in horizontal layout
    """
    if isinstance(font_color, str):
        font_color = vedo.getColor(font_color)
    sb = vtk.vtkScalarBarActor()
    sb.SetLabelFormat('%-#6.4g')
    #print(sb.GetLabelFormat())
    sb.SetLookupTable(lut)
    sb.SetUseOpacity(use_alpha)
    sb.SetDrawFrame(0)
    sb.SetDrawBackground(0)
    if lut.GetUseBelowRangeColor():
        sb.DrawBelowRangeSwatchOn()
        sb.SetBelowRangeAnnotation('')
    if lut.GetUseAboveRangeColor():
        sb.DrawAboveRangeSwatchOn()
        sb.SetAboveRangeAnnotation('')
    if lut.GetNanColor() != (0.5, 0.0, 0.0, 1.0):
        sb.DrawNanAnnotationOn()
        sb.SetNanAnnotation('nan')

    if title:
        if "\\" in repr(title):
            for r in vedo.shapes._reps:
                title = title.replace(r[0], r[1])
        titprop = sb.GetTitleTextProperty()
        titprop.BoldOn()
        titprop.ItalicOff()
        titprop.ShadowOff()
        titprop.SetColor(font_color)
        titprop.SetVerticalJustificationToTop()
        titprop.SetFontSize(titleFontSize)
        titprop.SetFontFamily(vtk.VTK_FONT_FILE)
        titprop.SetFontFile(vedo.settings.fonts_path + vedo.settings.defaultFont +'.ttf')
        sb.SetTitle(title)
        sb.SetVerticalTitleSeparation(titleYOffset)
        sb.SetTitleTextProperty(titprop)

    sb.UnconstrainedFontSizeOn()
    sb.DrawAnnotationsOn()
    sb.DrawTickLabelsOn()
    sb.SetMaximumNumberOfColors(256)

    if horizontal:
        sb.SetOrientationToHorizontal()
        sb.SetNumberOfLabels(3)
        sb.SetTextPositionToSucceedScalarBar()
        sb.SetPosition(pos)
        sb.SetMaximumWidthInPixels(1000)
        sb.SetMaximumHeightInPixels(50)
    else:
        sb.SetNumberOfLabels(7)
        sb.SetTextPositionToPrecedeScalarBar()
        sb.SetPosition(pos[0]+0.09, pos[1])
        sb.SetMaximumWidthInPixels(60)
        sb.SetMaximumHeightInPixels(250)

    if size[0] is not None: sb.SetMaximumWidthInPixels(size[0])
    if size[1] is not None: sb.SetMaximumHeightInPixels(size[1])

    if nlabels is not None:
        sb.SetNumberOfLabels(nlabels)

    sctxt = sb.GetLabelTextProperty()
    sctxt.SetFontFamily(vtk.VTK_FONT_FILE)
    sctxt.SetFontFile(vedo.settings.fonts_path + vedo.settings.defaultFont +'.ttf')
    sctxt.SetColor(font_color)
    sctxt.ItalicOff()
    sctxt.SetShadow(0)
    sctxt.SetFontSize(titleFontSize-2)
    sb.SetAnnotationTextProperty(sctxt)
    sb.PickableOff()
    return sb

def add_caption_symbol(point, size=0.5, color='red', alpha=1.0, overlay_2d=True):
    """
    Add a 2D or 3D overlay (aka caption in VTK).
    Modified from vedo caption() method
    """
    #c = np.array(self.GetProperty().GetColor())/2
    color = vedo.colors.getColor(color)

    """
    if point is None:
        x0,x1,y0,y1,z0,z1 = self.GetBounds()
        pt = [(x0+x1)/2, (y0+y1)/2, z1]
        point = self.closestPoint(pt)
    """

    caption = vtk.vtkCaptionActor2D()
    caption.SetAttachmentPoint(point)
    caption.SetBorder(False)
    caption.SetLeader(True)
    sph = vtk.vtkSphereSource()
    sph.Update()
    caption.SetLeaderGlyphData(sph.GetOutput())
    caption.SetLeaderGlyphSize(5)
    caption.SetMaximumLeaderGlyphSize(5)
    #capt.SetPadding(pad)
    #capt.SetCaption(txt)
    #capt.SetWidth(width)
    #capt.SetHeight(height)
    caption.SetThreeDimensionalLeader(not overlay_2d)

    prop = caption.GetProperty()
    prop.SetColor(color)
    prop.SetOpacity(alpha)
    """
    pr = caption.GetCaptionTextProperty()
    pr.SetFontFamily(vtk.VTK_FONT_FILE)
    if 'LogoType' in font: # special case of big file
        fl = vedo.io.download("https://vedo.embl.es/fonts/LogoType.ttf")
    else:
        fl = settings.fonts_path + font + '.ttf'
    if not os.path.isfile(fl):
        fl = font
    pr.SetFontFile(fl)
    pr.ShadowOff()
    pr.BoldOff()
    pr.FrameOff()
    pr.SetColor(c)
    pr.SetOpacity(alpha)
    pr.SetJustificationToLeft()
    if "top" in justify:
        pr.SetVerticalJustificationToTop()
    if "bottom" in justify:
        pr.SetVerticalJustificationToBottom()
    if "cent" in justify:
        pr.SetVerticalJustificationToCentered()
        pr.SetJustificationToCentered()
    if "left" in justify:
        pr.SetJustificationToLeft()
    if "right" in justify:
        pr.SetJustificationToRight()
    pr.SetLineSpacing(vspacing)
    self._caption = capt
    """
    return caption


def load_gltf_mesh(file_path='./data//brain_regions.glb'):
    """
    Read surface data from a binary GLTF scene
    """
    loader = vtk.vtkGLTFDocumentLoader()
    reader = vtk.vtkGLTFReader()
    reader.SetFileName(file_path)
    reader.Update() 
    #reader.Read()

    mb = reader.GetOutput()
    iterator = mb.NewIterator()

    vtk_polyobjects = []
    while not iterator.IsDoneWithTraversal():
        item = iterator.GetCurrentDataObject()
        vtk_polyobjects.append(item)
        iterator.GoToNextItem()

    print('Read', len(vtk_polyobjects), 'objects from glb')
    """ 
    output_port = reader.GetOutputPort()

    mapper = vtkPolyDataMapper()
    mapper.SetInputConnection(output_port)

    actor = vtkActor()
    actor.SetMapper(mapper) 
    """
    return vtk_polyobjects


import numpy as np
from vtk.util.numpy_support import numpy_to_vtk
import vtk
import vedo

"""
The classes are modifications of existing vedo classes that were either doing 
things in unsatisfying manner or lacking features.
"""


def Cross3D(pos=(0,0,0), size=1.0, thickness=0.5, color='black', alpha=1, res=4, lines_mode=True):
    """
    Build a 3D cross shape, mainly useful as a 3D marker.
    :param pos: Position of the cross
    :param size: Size of the cross
    :param thickness: Thickness in pixels (remains constant on screen)
    :param color: Color of the cross
    :param alpha: Alpha/opacity of the cross
    :param res: Resolution of the cylinders if not used in line_mode
    :param lines_mode: Whether lines are used or cylinders. The difference is in how they look like
        when you zoom in on the cross. If you use lines, their thickness is constant on screen, whether
        close or far. If you use cylinders, they will get thicker the closer you approach the camera
        (in perspective mode of course)
    :return: vedo.Mesh
    """
    if lines_mode:
        x1 = np.array([1.0, 0.0, 0.0]) * size / 2
        x2 = np.array([-1.0, 0.0, 0.0]) * size / 2
        c1 = vedo.Line(x1, x2, lw=thickness)

        y1 = np.array([0.0, 1.0, 0.0]) * size / 2
        y2 = np.array([0.0, -1.0, 0.0]) * size / 2
        c2 = vedo.Line(y1, y2, lw=thickness)

        z1 = np.array([0.0, 0.0, 1.0]) * size / 2
        z2 = np.array([0.0, 0.0, -1.0]) * size / 2
        c3 = vedo.Line(z1, z2, lw=thickness)
    else:
        c1 = vedo.Cylinder(r=thickness, height=size, res=res)
        c2 = vedo.Cylinder(r=thickness, height=size, res=res).rotateX(90)
        c3 = vedo.Cylinder(r=thickness, height=size, res=res).rotateY(90)
    cross = vedo.merge(c1,c2,c3).color(color).alpha(alpha)
    cross.SetPosition(pos)
    cross.name = 'Marker'
    return cross


class Lines(vedo.Lines):
    """
    Improved Lines class that supports point sets of varying lengths
    """
    def __init__(self, points, end_points=None, c='gray', alpha=1, lw=1, dotted=False):
        """
        Constructor
        parameters are the same as vedo.Line
        """
        self.axes = [1, 1, 1]
        #if not isinstance(point_sets, np.ndarray):
            #point_set = np.array(point_sets, dtype=object)
        if len(points.shape) > 1 and points.shape[1] == 2:
            super().__init__(points, end_points, c=c, alpha=alpha, lw=lw, dotted=dotted)
        else:
            polylns = vtk.vtkAppendPolyData()
            for point_set in points:
                positions = point_set
                if not isinstance(point_set, np.ndarray):
                    point_set = np.array(point_set)
                # numpy_to_vtk is unhappy if dtype is not to its liking
                point_set = point_set.astype(float)
                positions = numpy_to_vtk(np.ascontiguousarray(point_set), deep=True)

                # This part taken from class Line, which accepts n points
                vtk_points = vtk.vtkPoints()
                vtk_points.SetData(positions)
                
                lines = vtk.vtkCellArray()
                num_pts = len(point_set)
                lines.InsertNextCell(num_pts)
                for i in range(num_pts):
                    lines.InsertCellPoint(i)

                poly = vtk.vtkPolyData()
                poly.SetPoints(vtk_points)
                poly.SetLines(lines)
                polylns.AddInputData(poly)
            polylns.Update()

            vedo.Mesh.__init__(self, polylns.GetOutput())
            self.lw(lw).lighting('off')
            if dotted:
                self.GetProperty().SetLineStipplePattern(0xF0F0)
                self.GetProperty().SetLineStippleRepeatFactor(1)

        self.name = 'Lines'


class Points(vedo.Points):
    """
    Improved Points class that supports time series and screen-space mode
    """

    def __init__(self, positions, radius=1, values=None, color_map='viridis', screen_space=False, 
    alpha=1, res=6, min_v=None, max_v=None, scalars_prefix=None):
        """
        Constructor
        :param positions: 3D positions
        :param radius: Radius of the points
        :param values: Custom scalar values. You may pass a list the same length 
            as the number of points. If values is a 2D array, then we assume these are time series.
            You will then need to call:
            actor.polydata().GetPointData().SetActiveScalars(name)
            actor.mapper().SelectColorArray(name)
            where name is the array name that starts with the given scalars_prefix.
            If you have three steps in your time series, you will have by default
            Scalars_0, Scalars_1, Scalars_2.
        :param color_map: Color map, either a list of values and corresponding colors
            or a color map name (see vedo documentation with color maps that follow matplotlib)
        :param screen_space: Whether the points are rendered as screen-space points or as
            spheres. The main difference is that screen-space is very fast and can display millions
            of points whereas sphere mode allows you to zoom in on a point and you will see it
            bigger up-close with a perspective camera, but if you get too close, you will see that 
            the sphere is actually made of several scaled screen-space points.
        :param alpha: Alpha/opacity value
        :param res: Resolution of the point if screen-space is disabled
        :param min_v: Minimum value for the given values (will be computed if not given)
        :param max_v: Maximum value for the given values (will be computed if not given)
        :param scalars_prefix: Scalar array name prefix. The rest of the name is _id
            where id starts at 0.
        """
        self.scalars_prefix = 'Scalars_' if scalars_prefix is None else scalars_prefix
        self.axes = [1, 1, 1]

        # Multi component (ndimensional) arrays in vtk:
        # https://vtk.org/doc/nightly/html/classvtkAbstractArray.html#a528de7a4879a219e7f82a82130186dc8
        polydata = vtk.vtkPolyData()
        points = vtk.vtkPointSource()
        num_points = len(positions)
        points.SetNumberOfPoints(num_points)
        points.Update()
        #for p_id in range(num_points):
            #points.SetPoint(p_id, positions[p_id])
        if not isinstance(positions, np.ndarray):
            positions = np.array(positions)
        positions = positions.astype(float)
        points_data = numpy_to_vtk(np.ascontiguousarray(positions), deep=True)
        #polydata.SetPoints(points)
        polydata = points.GetOutput()
        polydata.GetPoints().SetData(points_data)

        # We have to set scalar values after the object is created because VTK automatically
        # creates some array values initially and we want to ignore them later on
        scalars = []
        if values is not None and len(values) > 0:
            if not isinstance(values, np.ndarray):
                values = np.array(values)
            if len(values.shape) > 1 and values.shape[1] > 1:
                # Safeguard
                if num_points == values.shape[0]:
                    values.reshape(num_points, -1)
                all_values = values.ravel()
                if min_v is None:
                    min_v = min(all_values)
                if max_v is None:
                    max_v = max(all_values)
                for loop_id in range(values.shape[1]):
                    scalars.append(self.add_scalar_data(polydata, values[:, loop_id]))
            else:
                if min_v is None:
                    min_v = min(values)
                if max_v is None:
                    max_v = max(values)
                scalars.append(self.add_scalar_data(polydata, values))

        # The following sets the "ActiveScalar", i.e. default scalar used by
        # VTK (and is must for setting radius of each sphere by corresponding
        # value in radii array). But what tells VTK to use it for scaling?
        ctf = None
        values_range = None
        if len(scalars) > 0:
            polydata.GetPointData().SetActiveScalars(scalars[0].GetName())
            if isinstance(color_map, vtk.vtkColorTransferFunction):
                ctf = color_map
            elif isinstance(color_map, vtk.vtkLookupTable):
                ctf = color_map
            elif isinstance(color_map, list) or isinstance(color_map, np.ndarray):
                ctf = vtk.vtkColorTransferFunction()
                for entry in color_map:
                    ctf.AddRGBPoint(entry[0], *entry[1])
            else:
                ctf = vtk.vtkColorTransferFunction()
                values_range = np.linspace(min_v, max_v, 20)
                for v_id in range(len(values_range)):
                    value = values_range[v_id]
                    ctf.AddRGBPoint(value, *vedo.colorMap(value, color_map, min_v, max_v))

        if screen_space:
            glyph = vtk.vtkVertexGlyphFilter()
            glyph.SetInputData(polydata)
        else:
            sphere = vtk.vtkSphereSource()
            sphere.SetRadius(radius)
            sphere.SetThetaResolution(res)
            sphere.SetPhiResolution(res)
            sphere.Update()

            glyph = vtk.vtkGlyph3D()
            glyph.SetSourceConnection(sphere.GetOutputPort())
            glyph.SetInputData(polydata)
            
            #glyph.SetVectorModeToUseVector()
            #glyph.OrientOn()
            glyph.ClampingOn()
            if len(scalars) > 0:
                # If a value == min_v, then the point has radius 0
                glyph.SetRange(min_v - 1, max_v)
            
            glyph.ScalingOn()
            glyph.SetScaleFactor(1)
            glyph.SetScaleModeToScaleByScalar()
            
            glyph.SetScaleMode(3)
            
            '''
            For further work:

            # Tell glyph which attribute arrays to use for what
            glyph.SetInputArrayToProcess(0, 0, 0, 0, 'Elevation')		# scalars
            #glyph.SetInputArrayToProcess(1,0,0,0,'RTDataGradient')		# vectors
            # glyph.SetInputArrayToProcess(2,0,0,0,'nothing')		# normals
            glyph.SetInputArrayToProcess(3, 0, 0, 0, 'RTData')		# colors

            # Calling update because I'm going to use the scalar range to set the color map range
            glyph.Update()

            coloring_by = 'RTData'
            mapper = vtk.vtkPolyDataMapper()
            mapper.SetInputConnection(glyph.GetOutputPort())
            mapper.SetScalarModeToUsePointFieldData()
            mapper.SetColorModeToMapScalars()
            mapper.ScalarVisibilityOn()
            mapper.SetScalarRange(glyph.GetOutputDataObject(0).GetPointData().GetArray(coloring_by).GetRange())
            mapper.SelectColorArray(coloring_by)
            actor = vtk.vtkActor()
            actor.SetMapper(mapper)
            '''

        glyph.Update()
        vedo.Mesh.__init__(self, glyph.GetOutput(), alpha=alpha)
        mapper = self._mapper

        self._polydata = polydata
        self.source = glyph
        self.name = 'Points'

        if screen_space:
            #self.GetProperty().SetColor(*color)
            self.GetProperty().SetPointSize(radius)
            self.GetProperty().SetRenderPointsAsSpheres(True)
        else:
            mapper.SetScalarModeToUsePointFieldData()
            if len(scalars) > 0:
                mapper.SetScalarRange(min_v, max_v)
            mapper.SetColorModeToMapScalars()

        if len(scalars) > 0:
            mapper.SelectColorArray(scalars[0].GetName())
        if ctf is not None:
            mapper.SetLookupTable(ctf)
        mapper.Update()

    def get_number_of_arrays(self, ignore=['GlyphScale', 'Normals']):
        """
        """
        num_scalar_arrays = 0
        point_data = self._polydata.GetPointData()
        for i in range(point_data.GetNumberOfArrays()):
            if point_data.GetArrayName(i) in ignore:
                pass
            num_scalar_arrays += 1
        return num_scalar_arrays

    def add_scalar_data(self, polydata, values, step_id=None):
        """
        Add scalar data to a VTK polydata
        :param polydata: vtkPolyData
        :param values: Numpy 1D array or list
        :param step_id: ID for naming the scalar with Scalars_#ID
        :return: vtkFloatArray
        """
        scalars = numpy_to_vtk(np.ascontiguousarray(values), deep=True)
        num_existing_ones = polydata.GetPointData().GetNumberOfArrays()
        if step_id is None:
            step_id = num_existing_ones
        scalars.SetName(self.scalars_prefix + str(step_id))
        polydata.GetPointData().AddArray(scalars)
        return scalars

    def update_data(self, positions, scalars=None):
        """
        Update the positions of points and optionally their scalar values
        :param positions: 3D coordinates the same length as the number 
        of points in the object
        :param scalars: 1D list or numpy array the same length as positions
        """
        vtk_positions = numpy_to_vtk(np.ascontiguousarray(positions), deep=True)
        polydata = self.source.GetInput()
        polydata.GetPoints().SetData(vtk_positions)
        if scalars is not None:
            #sn = polydata.GetPointData().GetScalars().GetName()
            for v_id in range(len(scalars)):
                polydata.GetPointData().GetScalars().SetValue(v_id, scalars[v_id])
        self.source.Update()


class Spheres(vedo.Mesh):
    """
    Reimplementation of vedo.Spheres that was not handling things properly
    when it comes to setting time series and to visualising them with colors.

    This class isn't used at the moment. Points is preferred as it acts either as
    vedo.Spheres or as screen space Points depending on screen_space param.

    In general, vedo uses "c" and "r" short variable names which are a hindrance...
    utils.Spheres is deprecated and you should favor utils.Points instead.
    """
    def __init__(self, centers, r=1, c="r", alpha=1, res=8):
        """
        Constructor.
        Parameters are the same as vedo.Spheres
        """
        self.axes = [1, 1, 1]
        if isinstance(centers, vedo.Points):
            centers = centers.points()

        cisseq = False
        if vedo.utils.isSequence(c):
            cisseq = True

        if cisseq:
            if len(centers) > len(c):
                vedo.printc("\times Mismatch in Spheres() colors", len(centers), len(c), c='r')
                raise RuntimeError()
            if len(centers) != len(c):
                vedo.printc("\lightningWarning: mismatch in Spheres() colors", len(centers), len(c))

        risseq = False
        if vedo.utils.isSequence(r):
            risseq = True

        if risseq:
            if len(centers) > len(r):
                vedo.printc("times Mismatch in Spheres() radius", len(centers), len(r), c='r')
                raise RuntimeError()
            if len(centers) != len(r):
                vedo.printc("\lightning Warning: mismatch in Spheres() radius", len(centers), len(r))
        if cisseq and risseq:
            vedo.printc("\noentry Limitation: c and r cannot be both sequences.", c='r')
            raise RuntimeError()

        src = vtk.vtkSphereSource()
        if not risseq:
            src.SetRadius(r)
        if vedo.utils.isSequence(res):
            res_t, res_phi = res
        else:
            res_t, res_phi = 2*res, res

        src.SetThetaResolution(res_t)
        src.SetPhiResolution(res_phi)
        src.Update()

        psrc = vtk.vtkPointSource()
        psrc.SetNumberOfPoints(len(centers))
        psrc.Update()
        pd = psrc.GetOutput()
        vpts = pd.GetPoints()

        glyph = vtk.vtkGlyph3D()
        glyph.SetSourceConnection(src.GetOutputPort())

        if cisseq:
            glyph.SetColorModeToColorByScalar()
            ucols = vtk.vtkUnsignedCharArray()
            ucols.SetNumberOfComponents(3)
            ucols.SetName("colors")
            #for i, p in enumerate(centers):
            for cx, cy, cz in c:
                #cx, cy, cz = getColor(acol)
                ucols.InsertNextTuple3(cx * 255, cy * 255, cz * 255)
            pd.GetPointData().SetScalars(ucols)
            glyph.ScalingOff()
        elif risseq:
            glyph.SetScaleModeToScaleByScalar()
            urads = numpy_to_vtk(np.ascontiguousarray(2*r).astype(float), deep=True)
            urads.SetName("radii")
            pd.GetPointData().SetScalars(urads)

        vpts.SetData(numpy_to_vtk(np.ascontiguousarray(centers), deep=True))

        glyph.SetInputData(pd)
        glyph.Update()

        vedo.Mesh.__init__(self, glyph.GetOutput(), alpha=alpha)
        self.phong()

        self._polydata = pd

        if cisseq:
            self.mapper().ScalarVisibilityOn()
        else:
            self.mapper().ScalarVisibilityOff()
            self.GetProperty().SetColor(vedo.getColor(c))
        self.name = 'Spheres'
